#! /usr/bin/env python
"""
utils - General tools of use to fileseq operations.
"""

import os
from itertools import chain, count, islice

from fileseq import exceptions 


def lenRange(start, stop, step=1):
    """
    Get the length of values for a given range
    """
    return (stop - start + step - 1 + 2 * (step < 0)) // step


class xrange2(object):
    """
    An itertools-based replacement for xrange which does
    not exhibit the OverflowError issue on some platforms, 
    when a value exceeds a C long size.
    
    Provides the features of an islice, with the added support
    for checking the length of the range.
    """
    
    __slots__ = ['_len', '_islice']
    
    def __init__(self, start, stop=None, step=1):
        if stop is None:
            start, stop = 0, start

        self._len = lenRange(start, stop, step)
        self._islice = islice(count(start, step), self._len)
    
    def __len__(self):
        return self._len 
    
    def next(self):
        return self._islice.next()
    
    def __iter__(self):
        return self._islice.__iter__()
    

# Issue #44
# On Windows platform, it is possible for xrange to get an
# OverflowError if a value passed to xrange exceeds the size of a C long. 
# Switch to an alternate implementation.
if os.name == 'nt':
    xrange = xrange2
else:
    xrange = xrange


def xfrange(start, stop, step=1, maxSize=-1):
    """
    Returns a generator that yields the frames from start to stop, inclusive.
    In other words it adds or subtracts a frame, as necessary, to return the
    stop value as well, if the stepped range would touch that value.

    :type start: int
    :type stop: int
    :type step: int
    :param step: Note that the sign will be ignored
    :type maxSize: int
    :param maxSize: If >= raise a :class:`fileseq.exceptions.MaxSizeException` 
                    if size is exceeded
    :rtype: generator
    """
    if start <= stop:
        stop, step = stop + 1, abs(step)
    else:
        stop, step = stop - 1, -abs(step)
        
    if maxSize >= 0:
        size = lenRange(start, stop, step)
        if size > maxSize:
            raise exceptions.MaxSizeException(
                "Size %d > %s (MAX_FRAME_SIZE)" % (size, maxSize))
        
    # because an xrange is an odd object all its own, we wrap it in a
    # generator expression to get a proper Generator
    return (f for f in xrange(start, stop, step))

def unique(seen, *iterables):
    """
    Get the unique items in iterables while preserving order.  Note that this
    mutates the seen set provided only when the returned generator is used.

    :param seen: either an empty set, or the set of things already seen
    :param iterables: one or more iterable lists to chain together
    :rtype: generator
    """
    _add = seen.add
    # return a generator of the unique items and the set of the seen items
    # the seen set will mutate when the generator is iterated over
    return (i for i in chain(*iterables) if i not in seen and not _add(i))

def pad(number, width=0):
    """
    Return the zero-padded string of a given number.

    :type number: int
    :param number: the number to pad
    :type width: int
    :param width: width for zero padding
    :rtype: str
    """
    return str(number).zfill(width)

def _getPathSep(path):
    """
    Abstracts returning the appropriate path separator 
    for the given path string. 

    This implementation always returns ``os.sep``

    Abstracted to make test mocking easier.

    :type path: str
    :param path: A path to check for the most common sep
    :rtype: str
    """
    return os.sep



#! /usr/bin/env python
"""
frameset - A set-like object representing a frame range for fileseq.
"""

import numbers

from collections import Set, Sequence

from fileseq import constants
from fileseq.constants import PAD_MAP, FRANGE_RE, PAD_RE
from fileseq.exceptions import MaxSizeException, ParseException
from fileseq.utils import xfrange, unique, pad

# Issue #44
# Possibly use an alternate xrange implementation, depending on platform. 
from fileseq.utils import xrange

class FrameSet(Set):
    """
    A :class:`FrameSet` is an immutable representation of the ordered, unique
    set of frames in a given frame range.

    The frame range can be expressed in the following ways:
        - 1-5
        - 1-5,10-20
        - 1-100x5 (every fifth frame)
        - 1-100y5 (opposite of above, fills in missing frames)
        - 1-100:4 (same as 1-100x4,1-100x3,1-100x2,1-100)

    A :class:`FrameSet` is effectively an ordered frozenset, with
    FrameSet-returning versions of frozenset methods:

        >>> FrameSet('1-5').union(FrameSet('5-10'))
        FrameSet('1-10')
        >>> FrameSet('1-5').intersection(FrameSet('5-10'))
        FrameSet('5')

    Because a FrameSet is hashable, it can be used as the key to a dictionary:

        >>> {FrameSet('1-20'): 'good'}

    Caveats:
        1. Because the internal storage of a ``FrameSet`` contains the discreet
           values of the entire range, an exception will be thrown if the range
           exceeds a large reasonable limit, which could lead to huge memory 
           allocations or memory failures. See `fileseq.constants.MAX_FRAME_SIZE`.
        2. All frozenset operations return a normalized :class:`FrameSet`:
           internal frames are in numerically increasing order.
        3. Equality is based on the contents and order, NOT the frame range
           string (there are a finite, but potentially
           extremely large, number of strings that can represent any given range,
           only a "best guess" can be made).
        4. Human-created frame ranges (ie 1-100x5) will be reduced to the
           actual internal frames (ie 1-96x5).
        5. The "null" :class:`Frameset` (``FrameSet('')``) is now a valid thing
           to create, it is required by set operations, but may cause confusion
           as both its start and end methods will raise IndexError.  The
           :meth:`is_null`
           property has been added to allow you to guard against this.

    :type frange: str
    :param frange: the frame range as a string (ie "1-100x5")
    :rtype: None
    :raises: :class:`fileseq.exceptions.ParseException` if the frame range
             (or a portion of it) could not be parsed.
             :class:`fileseq.exceptions.MaxSizeException` if the range exceeds
             `fileseq.constants.MAX_FRAME_SIZE`
    """

    __slots__ = ('_frange', '_items', '_order')

    def __new__(cls, *args, **kwargs):
        """
        Initialize the :class:`FrameSet` object.

        :type frange: str
        :param frange: the frame range as a string (ie "1-100x5")
        :returns: the :class:`FrameSet` instance
        :raises: :class:`fileseq.exceptions.ParseException` if the frame range
                 (or a portion of it) could not be parsed
        """
        self = super(cls, FrameSet).__new__(cls, *args, **kwargs)
        return self


    def __init__(self, frange):        
        # if the user provides anything but a string, short-circuit the build
        if not isinstance(frange, basestring):
            # if it's apparently a FrameSet already, short-circuit the build
            if set(dir(frange)).issuperset(self.__slots__):
                for attr in self.__slots__:
                    setattr(self, attr, getattr(frange, attr))
                return
            # if it's inherently disordered, sort and build
            elif isinstance(frange, Set):
                self._maxSizeCheck(frange)
                self._items = frozenset(map(int, frange))
                self._order = tuple(sorted(self._items))
                self._frange = FrameSet.framesToFrameRange(
                    self._order, sort=False, compress=False)
                return
            # if it's ordered, find unique and build
            elif isinstance(frange, Sequence):
                self._maxSizeCheck(frange)
                items = set()
                order = unique(items, map(int, frange))
                self._order = tuple(order)
                self._items = frozenset(items)
                self._frange = FrameSet.framesToFrameRange(
                    self._order, sort=False, compress=False)
                return
            # in all other cases, cast to a string
            else:
                try:
                    frange = str(frange)
                except Exception as err:
                    msg = 'Could not parse "{0}": cast to string raised: {1}'
                    raise ParseException(msg.format(frange, err))

        # we're willing to trim padding characters from consideration
        # this translation is orders of magnitude faster than prior method
        self._frange = str(frange).translate(None, ''.join(PAD_MAP.keys()))

        # because we're acting like a set, we need to support the empty set
        if not self._frange:
            self._items = frozenset()
            self._order = tuple()
            return

        # build the mutable stores, then cast to immutable for storage
        items = set()
        order = []

        maxSize = constants.MAX_FRAME_SIZE 

        for part in self._frange.split(","):
            # this is to deal with leading / trailing commas
            if not part:
                continue
            # parse the partial range
            start, end, modifier, chunk = FrameSet._parse_frange_part(part)
            # handle batched frames (1-100x5)
            if modifier == 'x':
                frames = xfrange(start, end, chunk, maxSize=maxSize)
                frames = [f for f in frames if f not in items]
                self._maxSizeCheck(len(frames) + len(items))
                order.extend(frames)
                items.update(frames)
            # handle staggered frames (1-100:5)
            elif modifier == ':':
                for stagger in xrange(chunk, 0, -1):
                    frames = xfrange(start, end, stagger, maxSize=maxSize)
                    frames = [f for f in frames if f not in items]
                    self._maxSizeCheck(len(frames) + len(items))
                    order.extend(frames)
                    items.update(frames)
            # handle filled frames (1-100y5)
            elif modifier == 'y':
                not_good = frozenset(xfrange(start, end, chunk, maxSize=maxSize))
                frames = xfrange(start, end, 1, maxSize=maxSize)
                frames = (f for f in frames if f not in not_good)
                frames = [f for f in frames if f not in items]
                self._maxSizeCheck(len(frames) + len(items))
                order.extend(frames)
                items.update(frames)
            # handle full ranges and single frames
            else:
                frames = xfrange(start, end, 1 if start < end else -1, maxSize=maxSize)
                frames = [f for f in frames if f not in items]
                self._maxSizeCheck(len(frames) + len(items))
                order.extend(frames)
                items.update(frames)

        # lock the results into immutable internals
        # this allows for hashing and fast equality checking
        self._items = frozenset(items)
        self._order = tuple(order)

    @property
    def is_null(self):
        """
        Read-only access to determine if the :class:`FrameSet` is the null or
        empty :class:`FrameSet`.

        :rtype: bool
        """
        return not (self._frange and self._items and self._order)

    @property
    def frange(self):
        """
        Read-only access to the frame range used to create this :class:`FrameSet`.

        :rtype: frozenset
        """
        return self._frange

    @property
    def items(self):
        """
        Read-only access to the unique frames that form this :class:`FrameSet`.

        :rtype: frozenset
        """
        return self._items

    @property
    def order(self):
        """
        Read-only access to the ordered frames that form this :class:`FrameSet`.

        :rtype: tuple
        """
        return self._order

    @classmethod
    def from_iterable(cls, frames, sort=False):
        """
        Build a :class:`FrameSet` from an iterable of frames.

        :param frames: an iterable object containing frames as integers
        :param sort: True to sort frames before creation, default is False
        :rtype: :class:`FrameSet`
        """
        return FrameSet(sorted(frames) if sort else frames)

    @classmethod
    def _cast_to_frameset(cls, other):
        """
        Private method to simplify comparison operations.

        :param other: the :class:`FrameSet`, set, frozenset, or iterable to be compared
        :rtype: :class:`FrameSet`
        :returns: :class:`NotImplemented` if a comparison is impossible
        """
        if isinstance(other, FrameSet):
            return other
        try:
            return FrameSet(other)
        except Exception:
            return NotImplemented

    def index(self, frame):
        """
        Return the index of the given frame number within the :class:`FrameSet`.

        :type frame: int
        :param frame: the frame number to find the index for
        :rtype: int
        :raises: :class:`ValueError` if frame is not in self
        """
        return self.order.index(frame)

    def frame(self, index):
        """
        Return the frame at the given index.

        :type index: int
        :param index: the index to find the frame for
        :rtype: int
        :raises: :class:`IndexError` if index is out of bounds
        """
        return self.order[index]

    def hasFrame(self, frame):
        """
        Check if the :class:`FrameSet` contains the frame.

        :type frame: int
        :param frame: the frame number to search for
        :rtype: bool
        """
        return frame in self

    def start(self):
        """
        The first frame in the :class:`FrameSet`.

        :rtype: int
        :raises: :class:`IndexError` (with the empty :class:`FrameSet`)
        """
        return self.order[0]

    def end(self):
        """
        The last frame in the :class:`FrameSet`.

        :rtype: int
        :raises: :class:`IndexError` (with the empty :class:`FrameSet`)
        """
        return self.order[-1]

    def frameRange(self, zfill=0):
        """
        Return the frame range used to create this :class:`FrameSet`, padded if
        desired.

        :Example:
            >>> FrameSet('1-100').frameRange()
            '1-100'
            >>> FrameSet('1-100').frameRange(5)
            '00001-00100'

        :type zfill: int
        :param zfill: the width to use to zero-pad the frame range string
        :rtype: str
        """
        return FrameSet.padFrameRange(self.frange, zfill)

    def invertedFrameRange(self, zfill=0):
        """
        Return the inverse of the :class:`FrameSet` 's frame range, padded if
        desired.
        The inverse is every frame within the full extent of the range.

        :Example:
            >>> FrameSet('1-100x2').invertedFrameRange()
            '2-98x2'
            >>> FrameSet('1-100x2').invertedFrameRange(5)
            '00002-00098x2'

        If the inverted frame size exceeds `fileseq.constants.MAX_FRAME_SIZE`, 
        a ``MaxSizeException`` will be raised.

        :type zfill: int
        :param zfill: the width to use to zero-pad the frame range string
        :rtype: str
        :raises: :class:`fileseq.exceptions.MaxSizeException`
        """
        result = []
        frames = sorted(self.items)
        for idx, frame in enumerate(frames[:-1]):
            next_frame = frames[idx + 1]
            if next_frame - frame != 1:
                r = xrange(frame + 1, next_frame)
                # Check if the next update to the result set
                # will exceed out max frame size. 
                # Prevent memory overflows.
                self._maxSizeCheck(len(r) + len(result))
                result += r
        
        if not result:
            return ''
        
        return FrameSet.framesToFrameRange(
            result, zfill=zfill, sort=False, compress=False)

    def normalize(self):
        """
        Returns a new normalized (sorted and compacted) :class:`FrameSet`.

        :rtype: :class:`FrameSet`
        """
        return FrameSet(FrameSet.framesToFrameRange(
            self.items, sort=True, compress=False))

    def __getstate__(self):
        """
        Allows for serialization to a pickled :class:`FrameSet`.

        :rtype: tuple (frame range string, )
        """
        # we have to special-case the empty FrameSet, because of a quirk in
        # Python where __setstate__ will not be called if the return value of
        # bool(__getstate__) == False.  A tuple with ('',) will return True.
        return (self.frange, )

    def __setstate__(self, state):
        """
        Allows for de-serialization from a pickled :class:`FrameSet`.

        :type state: tuple, str, or dict
        :param state: A string/dict can be used for backwards compatibility
        :rtype: None
        :raises: :class:`ValueError` if state is not an appropriate type
        """
        if isinstance(state, tuple):
            # this is to allow unpickling of "3rd generation" FrameSets,
            # which are immutable and may be empty.
            self.__init__(state[0])
        elif isinstance(state, basestring):
            # this is to allow unpickling of "2nd generation" FrameSets,
            # which were mutable and could not be empty.
            self.__init__(state)
        elif isinstance(state, dict):
            # this is to allow unpickling of "1st generation" FrameSets,
            # when the full __dict__ was stored
            if '__frange' in state and '__set' in state and '__list' in state:
                self._frange = state['__frange']
                self._items = frozenset(state['__set'])
                self._order = tuple(state['__list'])
            else:
                for k in self.__slots__:
                    setattr(self, k, state[k])
        else:
            msg = "Unrecognized state data from which to deserialize FrameSet"
            raise ValueError(msg)

    def __getitem__(self, index):
        """
        Allows indexing into the ordered frames of this :class:`FrameSet`.

        :type int:
        :param index: the index to retrieve
        :rtype: int
        :raises: :class:`IndexError` if index is out of bounds
        """
        return self.order[index]

    def __len__(self):
        """
        Returns the length of the ordered frames of this :class:`FrameSet`.

        :rtype: int
        """
        return len(self.order)

    def __str__(self):
        """
        Returns the frame range string of this :class:`FrameSet`.

        :rtype: str
        """
        return self.frange

    def __repr__(self):
        """
        Returns a long-form representation of this :class:`FrameSet`.

        :rtype: str
        """
        return '{0}("{1}")'.format(self.__class__.__name__, self.frange)

    def __iter__(self):
        """
        Allows for iteration over the ordered frames of this :class:`FrameSet`.

        :rtype: generator
        """
        return (i for i in self.order)

    def __reversed__(self):
        """
        Allows for reversed iteration over the ordered frames of this
        :class:`FrameSet`.

        :rtype: generator
        """
        return (i for i in reversed(self.order))

    def __contains__(self, item):
        """
        Check if item is a member of this :class:`FrameSet`.

        :type item: int
        :param item: the frame number to check for
        :rtype: bool
        """
        return item in self.items

    def __hash__(self):
        """
        Builds the hash of this :class:`FrameSet` for equality checking and to
        allow use as a dictionary key.

        :rtype: int
        """
        return hash(self.frange) | hash(self.items) | hash(self.order)

    def __lt__(self, other):
        """
        Check if self < other via a comparison of the contents. If other is not
        a :class:`FrameSet`, but is a set, frozenset, or is iterable, it will be
        cast to a :class:`FrameSet`.

        .. note::

            A :class:`FrameSet` is less than other if the set of its contents are
            less, OR if the contents are equal but the order of the items is less.

            .. code-block:: python
                :caption: Same contents, but (1,2,3,4,5) sorts below (5,4,3,2,1)

                >>> FrameSet("1-5") < FrameSet("5-1")
                True

        :type other: FrameSet
        :param other: Can also be an object that can be cast to a :class:`FrameSet`
        :rtype: bool, or :class:`NotImplemented` if `other` fails to convert
                to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.items < other.items or (
            self.items == other.items and self.order < other.order)

    def __le__(self, other):
        """
        Check if `self` <= `other` via a comparison of the contents.
        If `other` is not a :class:`FrameSet`, but is a set, frozenset, or
        is iterable, it will be cast to a :class:`FrameSet`.

        :type other: FrameSet
        :param other: Also accepts an object that can be cast to a :class:`FrameSet`
        :rtype: bool, or :class:`NotImplemented` if `other` fails to convert
                to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.items <= other.items

    def __eq__(self, other):
        """
        Check if `self` == `other` via a comparison of the hash of
        their contents.
        If `other` is not a :class:`FrameSet`, but is a set, frozenset, or
        is iterable, it will be cast to a :class:`FrameSet`.

        :type other: :class:`FrameSet`
        :param other: Also accepts an object that can be cast to a :class:`FrameSet`
        :rtype: bool, or :class:`NotImplemented` if `other` fails to convert
                to a :class:`FrameSet`
        """
        if not isinstance(other, FrameSet):
            if not hasattr(other, '__iter__'):
                return NotImplemented
            other = self.from_iterable(other)
        this = hash(self.items) | hash(self.order)
        that = hash(other.items) | hash(other.order)
        return this == that

    def __ne__(self, other):
        """
        Check if `self` != `other` via a comparison of the hash of
        their contents.
        If `other` is not a :class:`FrameSet`, but is a set, frozenset, or
        is iterable, it will be cast to a :class:`FrameSet`.

        :type other: :class:`FrameSet`
        :param other: Also accepts an object that can be cast to a :class:`FrameSet`
        :rtype: bool, or :class:`NotImplemented` if `other` fails to convert
                to a :class:`FrameSet`
        """
        is_equals = self == other
        if is_equals != NotImplemented:
            return not is_equals
        return is_equals

    def __ge__(self, other):
        """
        Check if `self` >= `other` via a comparison of the contents.
        If `other` is not a :class:`FrameSet`, but is a set, frozenset, or
        is iterable, it will be cast to a :class:`FrameSet`.

        :type other: :class:`FrameSet`
        :param other: Also accepts an object that can be cast to one a :class:`FrameSet`
        :rtype: bool, or :class:`NotImplemented` if `other` fails to convert
                to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.items >= other.items

    def __gt__(self, other):
        """
        Check if `self` > `other` via a comparison of the contents.
        If `other` is not a :class:`FrameSet`, but is a set, frozenset, or
        is iterable, it will be cast to a :class:`FrameSet`.

        .. note::
           A :class:`FrameSet` is greater than `other` if the set of its
           contents are greater,
           OR if the contents are equal but the order is greater.

           .. code-block:: python
               :caption: Same contents, but (1,2,3,4,5) sorts below (5,4,3,2,1)

               >>> FrameSet("1-5") > FrameSet("5-1")
               False

        :type other: :class:`FrameSet`
        :param other: Also accepts an object that can be cast to a :class:`FrameSet`
        :rtype: bool, or :class:`NotImplemented` if :param: other fails to convert
                to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.items > other.items or (
            self.items == other.items and self.order > other.order)

    def __and__(self, other):
        """
        Overloads the ``&`` operator.
        Returns a new :class:`FrameSet` that holds only the
        frames `self` and `other` have in common.

        .. note::

            The order of operations is irrelevant:
            ``(self & other) == (other & self)``

        :type other: :class:`FrameSet`
        :rtype: :class:`FrameSet`, or :class:`NotImplemented` if :param: other
                fails to convert to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.from_iterable(self.items & other.items, sort=True)

    __rand__ = __and__

    def __sub__(self, other):
        """
        Overloads the ``-`` operator.
        Returns a new :class:`FrameSet` that holds only the
        frames of `self` that are not in `other.`

        .. note::

            This is for left-hand subtraction (``self - other``).

        :type other: :class:`FrameSet`
        :rtype: :class:`FrameSet`, or :class:`NotImplemented` if `other`
                fails to convert to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.from_iterable(self.items - other.items, sort=True)

    def __rsub__(self, other):
        """
        Overloads the ``-`` operator.
        Returns a new :class:`FrameSet` that holds only the
        frames of `other` that are not in `self.`

        .. note::

            This is for right-hand subtraction (``other - self``).

        :type other: :class:`FrameSet`
        :rtype: :class:`FrameSet`, or :class:`NotImplemented` if `other`
                fails to convert to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.from_iterable(other.items - self.items, sort=True)

    def __or__(self, other):
        """
        Overloads the ``|`` operator.
        Returns a new :class:`FrameSet` that holds all the
        frames in `self,` `other,` or both.

        .. note::

            The order of operations is irrelevant:
            ``(self | other) == (other | self)``

        :type other: :class:`FrameSet`
        :rtype: :class:`FrameSet`, or :class:`NotImplemented` if `other`
                fails to convert to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.from_iterable(self.items | other.items, sort=True)

    __ror__ = __or__

    def __xor__(self, other):
        """
        Overloads the ``^`` operator.
        Returns a new :class:`FrameSet` that holds all the
        frames in `self` or `other` but not both.

        .. note::
            The order of operations is irrelevant:
            ``(self ^ other) == (other ^ self)``

        :type other: :class:`FrameSet`
        :rtype: :class:`FrameSet`, or :class:`NotImplemented` if `other`
                fails to convert to a :class:`FrameSet`.
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.from_iterable(self.items ^ other.items, sort=True)

    __rxor__ = __xor__

    def isdisjoint(self, other):
        """
        Check if the contents of :class:self has no common intersection with the
        contents of :class:other.

        :type other: :class:`FrameSet`
        :rtype: bool, or :class:`NotImplemented` if `other` fails to convert
                to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.items.isdisjoint(other.items)

    def issubset(self, other):
        """
        Check if the contents of `self` is a subset of the contents of
        `other.`

        :type other: :class:`FrameSet`
        :rtype: bool, or :class:`NotImplemented` if `other` fails to convert
                to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.items <= other.items

    def issuperset(self, other):
        """
        Check if the contents of `self` is a superset of the contents of
        `other.`

        :type other: :class:`FrameSet`
        :rtype: bool, or :class:`NotImplemented` if `other` fails to convert
                to a :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        return self.items >= other.items

    def union(self, *other):
        """
        Returns a new :class:`FrameSet` with the elements of `self` and
        of `other`.

        :type other: :class:`FrameSet` or objects that can cast to :class:`FrameSet`
        :rtype: :class:`FrameSet`
        """
        from_frozenset = self.items.union(*map(set, other))
        return self.from_iterable(from_frozenset, sort=True)

    def intersection(self, *other):
        """
        Returns a new :class:`FrameSet` with the elements common to `self` and
        `other`.

        :type other: :class:`FrameSet` or objects that can cast to :class:`FrameSet`
        :rtype: :class:`FrameSet`
        """
        from_frozenset = self.items.intersection(*map(set, other))
        return self.from_iterable(from_frozenset, sort=True)

    def difference(self, *other):
        """
        Returns a new :class:`FrameSet` with elements in `self` but not in
        `other`.

        :type other: :class:`FrameSet` or objects that can cast to :class:`FrameSet`
        :rtype: :class:`FrameSet`
        """
        from_frozenset = self.items.difference(*map(set, other))
        return self.from_iterable(from_frozenset, sort=True)

    def symmetric_difference(self, other):
        """
        Returns a new :class:`FrameSet` that contains all the elements in either
        `self` or `other`, but not both.

        :type other: :class:`FrameSet`
        :rtype: :class:`FrameSet`
        """
        other = self._cast_to_frameset(other)
        if other is NotImplemented:
            return NotImplemented
        from_frozenset = self.items.symmetric_difference(other.items)
        return self.from_iterable(from_frozenset, sort=True)

    def copy(self):
        """
        Returns a shallow copy of this :class:`FrameSet`.

        :rtype: :class:`FrameSet`
        """
        return FrameSet(str(self))

    @classmethod
    def _maxSizeCheck(cls, obj):
        """
        Raise a MaxSizeException if ``obj`` exceeds MAX_FRAME_SIZE

        :type obj: number or collection
        :raises: :class:`fileseq.exceptions.MaxSizeException`
        """
        fail = False
        size = 0

        if isinstance(obj, numbers.Number):
            if obj > constants.MAX_FRAME_SIZE:
                fail = True
                size = obj

        elif hasattr(obj, '__len__'):
            size = len(obj)
            fail = size > constants.MAX_FRAME_SIZE

        if fail:
            raise MaxSizeException('Frame size %s > %s (MAX_FRAME_SIZE)' \
                    % (size, constants.MAX_FRAME_SIZE))

    @staticmethod
    def isFrameRange(frange):
        """
        Return True if the given string is a frame range. Any padding
        characters, such as '#' and '@' are ignored.

        :type frange: str
        :param frange: a frame range to test
        :rtype: bool
        """
        # we're willing to trim padding characters from consideration
        # this translation is orders of magnitude faster than prior method
        frange = str(frange).translate(None, ''.join(PAD_MAP.keys()))
        if not frange:
            return True
        for part in frange.split(','):
            if not part:
                continue
            try:
                FrameSet._parse_frange_part(part)
            except ParseException:
                return False
        return True

    @staticmethod
    def padFrameRange(frange, zfill):
        """
        Return the zero-padded version of the frame range string.

        :type frange: str
        :param frange: a frame range to test
        :rtype: str
        """
        def _do_pad(match):
            """
            Substitutes padded for unpadded frames.
            """
            result = list(match.groups())
            result[1] = pad(result[1], zfill)
            if result[4]:
                result[4] = pad(result[4], zfill)
            return ''.join((i for i in result if i))
        return PAD_RE.sub(_do_pad, frange)

    @staticmethod
    def _parse_frange_part(frange):
        """
        Internal method: parse a discrete frame range part.

        :type frange: str
        :param frange: single part of a frame range as a string (ie "1-100x5")
        :rtype: tuple (start, end, modifier, chunk)
        :raises: :class:`fileseq.exceptions.ParseException` if the frame range
                 can not be parsed
        """
        match = FRANGE_RE.match(frange)
        if not match:
            msg = 'Could not parse "{0}": did not match {1}'
            raise ParseException(msg.format(frange, FRANGE_RE.pattern))
        start, end, modifier, chunk = match.groups()
        start = int(start)
        end = int(end) if end is not None else start
        chunk = abs(int(chunk)) if chunk is not None else 1
        # a zero chunk is just plain illogical
        if chunk == 0:
            msg = 'Could not parse "{0}": chunk cannot be 0'
            raise ParseException(msg.format(frange))
        return start, end, modifier, chunk

    @staticmethod
    def _build_frange_part(start, stop, stride, zfill=0):
        """
        Private method: builds a proper and padded
        :class:`fileseq.framerange.FrameRange` string.

        :type start: int
        :param start: first frame
        :type stop: int
        :param stop: last frame
        :type stride: int
        :param stride: increment
        :type zfill: int
        :param zfill: width for zero padding
        :rtype: str
        """
        if stop is None:
            return ''
        pad_start = pad(start, zfill)
        pad_stop = pad(stop, zfill)
        if stride is None or start == stop:
            return '{0}'.format(pad_start)
        elif abs(stride) == 1:
            return '{0}-{1}'.format(pad_start, pad_stop)
        else:
            return '{0}-{1}x{2}'.format(pad_start, pad_stop, stride)

    @staticmethod
    def framesToFrameRanges(frames, zfill=0):
        """
        Converts a sequence of frames to a series of padded
        :class:`fileseq.framerange.FrameRange` s.

        :type frames: iterable
        :param frames: sequence of frames to process
        :type zfill: int
        :param zfill: width for zero padding
        :rtype: generator
        """
        _build = FrameSet._build_frange_part
        curr_start = None
        curr_stride = None
        curr_frame = None
        last_frame = None
        curr_count = 0
        for curr_frame in frames:
            if curr_start is None:
                curr_start = curr_frame
                last_frame = curr_frame
                curr_count += 1
                continue
            if curr_stride is None:
                curr_stride = abs(curr_frame-curr_start)
            new_stride = abs(curr_frame-last_frame)
            if curr_stride == new_stride:
                last_frame = curr_frame
                curr_count += 1
            elif curr_count == 2:
                yield _build(curr_start, curr_start, None, zfill)
                curr_start = last_frame
                curr_stride = new_stride
                last_frame = curr_frame
            else:
                yield _build(curr_start, last_frame, curr_stride, zfill)
                curr_stride = None
                curr_start = curr_frame
                last_frame = curr_frame
                curr_count = 1
        if curr_count == 2:
            yield _build(curr_start, curr_start, None, zfill)
            yield _build(curr_frame, curr_frame, None, zfill)
        else:
            yield _build(curr_start, curr_frame, curr_stride, zfill)

    @staticmethod
    def framesToFrameRange(frames, sort=True, zfill=0, compress=False):
        """
        Converts an iterator of frames into a
        :class:`fileseq.framerange.FrameRange`.

        :type frames: iterable
        :param frames: sequence of frames to process
        :type sort: bool
        :param sort: sort the sequence before processing
        :type zfill: int
        :param zfill: width for zero padding
        :type compress: bool
        :param compress: remove any duplicates before processing
        :rtype: str
        """
        if compress:
            frames = unique(set(), frames)
        frames = list(frames)
        if not frames:
            return ''
        if len(frames) == 1:
            return pad(frames[0], zfill)
        if sort:
            frames.sort()
        return ','.join(FrameSet.framesToFrameRanges(frames, zfill))



#! /usr/bin/env python
"""
filesequence - A parsing object representing sequential files for fileseq.
"""

import os
from glob import iglob
from itertools import imap, ifilter
from fileseq.exceptions import ParseException, FileSeqException
from fileseq.constants import PAD_MAP, DISK_RE, SPLIT_RE, PRINTF_SYNTAX_PADDING_RE
from fileseq.frameset import FrameSet
from fileseq import utils

class FileSequence(object):
    """:class:`FileSequence` represents an ordered sequence of files.
    
    If the frame size exceeds :obj:`fileseq.constants.MAX_FRAME_SIZE`, 
    a ``MaxSizeException`` will be thrown.

    :type sequence: str
    :param sequence: (ie: dir/path.1-100#.ext)
    :raises: :class:`fileseq.exceptions.MaxSizeException`
    """
    def __init__(self, sequence):
        if not hasattr(self, '_frameSet'):

            self._frameSet = None

            try:
                # the main case, padding characters in the path.1-100#.exr
                path, frames, self._pad, self._ext = SPLIT_RE.split(sequence, 1)
                self._dir, self._base = os.path.split(path)
                self._frameSet = FrameSet(frames)
            except ValueError:
                # edge case 1; we've got an invalid pad
                for placeholder in PAD_MAP.keys():
                    if placeholder in sequence:
                        msg = "Failed to parse FileSequence: {0}"
                        raise ParseException(msg.format(sequence))
                # edge case 2; we've got a single frame of a sequence
                a_frame = DISK_RE.match(sequence)
                if a_frame:
                    self._dir, self._base, frames, self._ext = a_frame.groups()
                    # edge case 3: we've got a single versioned file, not a sequence
                    if frames and not self._base.endswith('.'):
                        self._base = self._base + frames
                        self._pad = ''
                    elif not frames:
                        self._pad = ''
                        self._frameSet = None
                    else:
                        self._frameSet = FrameSet(frames)
                        if self._frameSet:
                            self._pad = FileSequence.getPaddingChars(len(frames))
                        else:
                            self._pad = ''
                            self._frameSet = None
                # edge case 4; we've got a solitary file, not a sequence
                else:
                    path, self._ext = os.path.splitext(sequence)
                    self._dir, self._base = os.path.split(path)
                    self._pad = ''

        if self._dir:
            sep = utils._getPathSep(sequence)
            if not self._dir.endswith(sep):
                self._dir += sep

        self._zfill = self.__class__.getPaddingNum(self._pad)

    def format(self, template="{basename}{range}{padding}{extension}"):
        """Return the file sequence as a formatted string according to
        the given template.

        Utilizes the python string format syntax.  Available keys include:
            * basename - the basename of the sequence.
            * extension - the file extension of the sequence.
            * start - the start frame.
            * end - the end frame.
            * length - the length of the frame range.
            * padding - the detecting amount of padding.
            * inverted - the inverted frame range. (returns "" if none)
            * dirname - the directory name.
        
        If asking for the inverted range value, and the new inverted range
        exceeded :obj:`fileseq.constants.MAX_FRAME_SIZE`, a ``MaxSizeException``
        will be raised.

        :type template: str
        :rtype: str
        :raises: :class:`fileseq.exceptions.MaxSizeException` 
        """
        # Potentially expensive if inverted range is large
        # and user never asked for it in template
        inverted = (self.invertedFrameRange() or "") if "{inverted}" in template else ""

        return template.format(
            basename=self.basename(),
            extension=self.extension(), start=self.start(),
            end=self.end(), length=len(self),
            padding=self.padding(),
            range=self.frameRange() or "",
            inverted=inverted,
            dirname=self.dirname())

    def split(self):
        """
        Split the :class:`FileSequence` into contiguous pieces and return them
        as a list of :class:`FileSequence` instances.

        :rtype: list
        """
        result = []
        for frange in self.frameRange().split(","):
            result.append(FileSequence(''.join(
                (self._dir, self._base, frange, self._pad, self._ext))))
        return result

    def dirname(self):
        """
        Return the directory name of the sequence.

        :rtype: str
        """
        return self._dir

    def setDirname(self, dirname):
        """
        Set a new directory name for the sequence.

        :type dirname: str
        :param dirname: the new directory name
        :rtype: None
        """
        self._dir = dirname

    def basename(self):
        """
        Return the basename of the sequence.

        :rtype: str
        """
        return self._base

    def setBasename(self, base):
        """
        Set a new basename for the sequence.

        :type base: str
        :param base: the new base name
        :rtype: None
        """
        self._base = base

    def padding(self):
        """
        Return the the padding characters in the sequence.

        :rtype: str
        """
        return self._pad

    def setPadding(self, padding):
        """
        Set new padding characters for the sequence.
        i.e. "#" or "@@@" or '%04d', or an empty string to disable range formatting.

        :type padding: str
        :rtype: None
        """
        self._pad = padding
        self._zfill = self.__class__.getPaddingNum(self._pad)

    def frameSet(self):
        """
        Return the :class:`fileseq.frameset.FrameSet` of the sequence if specified,
        otherwise None.

        :rtype: :class:`fileseq.frameset.FrameSet` or None
        """
        return self._frameSet

    def setFrameSet(self, frameSet):
        """
        Set a new :class:`fileseq.frameset.FrameSet` for the sequence.

        :param frameSet: the new :class:`fileseq.frameset.FrameSet` object
        :rtype: None
        """
        self._frameSet = frameSet

    def extension(self):
        """
        Return the file extension of the sequence, including leading period.

        :rtype: str
        """
        return self._ext

    def setExtension(self, ext):
        """
        Set a new file extension for the sequence.

        .. note::
            A leading period will be added if none is provided.

        :param ext: the new file extension
        :rtype: None
        """
        if ext[0] != ".":
            ext = "." + ext
        self._ext = ext

    def setExtention(self, ext):
        """
        Deprecated: use :meth:`setExtension`.
        """
        import warnings
        msg = "the setExtention method is deprecated, please use setExtension"
        warnings.warn(msg)
        self.setExtension(ext)

    def frameRange(self):
        """
        Returns the string formatted frame range of the sequence.
        Will return an empty string if the sequence has no frame pattern.

        :rtype: str
        """
        if not self._frameSet:
            return ''
        return self._frameSet.frameRange(self._zfill)

    def setFrameRange(self, frange):
        """
        Set a new frame range for the sequence.

        :param frange: a properly formatted frame range, as per :class:`fileseq.frameset.FrameSet`
        :rtype: None
        """
        self._frameSet = FrameSet(frange)

    def invertedFrameRange(self):
        """
        Returns the inverse string formatted frame range of the sequence.
        Will return an empty string if the sequence has no frame pattern.

        If new inverted range exceeded :obj:`fileseq.constants.MAX_FRAME_SIZE`, 
        a ``MaxSizeException`` will be raised.

        :rtype: str
        :raises: :class:`fileseq.exceptions.MaxSizeException`
        """
        if not self._frameSet:
            return ''
        return self._frameSet.invertedFrameRange(self._zfill)

    def start(self):
        """
        Returns the start frame of the sequence's :class:`fileseq.frameset.FrameSet`.
        Will return 0 if the sequence has no frame pattern.

        :rtype: int
        """
        if not self._frameSet:
            return 0
        return self._frameSet.start()

    def end(self):
        """
        Returns the end frame of the sequences :class:`fileseq.frameset.FrameSet`.
        Will return 0 if the sequence has no frame pattern.

        :rtype: int
        """
        if not self._frameSet:
            return 0
        return self._frameSet.end()

    def zfill(self):
        """
        Returns the zfill depth (ie the number of zeroes to pad with).

        :rtype: int
        """
        return self._zfill

    def frame(self, frame):
        """
        Return a path go the given frame in the sequence.  Integer or string
        digits are treated as a frame number and padding is applied, all other
        values are passed though.

        :Example:
                >>> seq.frame(1)
                /foo/bar.0001.exr
                >>> seq.frame("#")
                /foo/bar.#.exr

        :param frame: the desired frame number (int/str) or a char to pass through (ie. #)
        :rtype: str
        """
        try:
            zframe = str(int(frame)).zfill(self._zfill)
        except ValueError:
            zframe = frame

        # There may have been no placeholder for frame IDs in
        # the sequence, in which case we don't want to insert
        # a frame ID

        if self._zfill == 0:
            zframe = ""

        return "".join((self._dir, self._base, zframe, self._ext))

    def index(self, idx):
        """
        Return the path to the file at the given index.

        :type idx: int
        :param idx: the desired index
        :rtype: str
        """
        if not self._frameSet:
            return str(self)
        return self.frame(self._frameSet[idx])

    def __iter__(self):
        """
        Allow iteration over the path or paths this :class:`FileSequence`
        represents.

        :rtype: generator
        """
        # If there is no frame range, or there is no padding
        # characters, then we only want to represent a single path
        if not self._frameSet or not self._zfill:
            yield str(self)
            return

        for f in self._frameSet:
            yield self.frame(f)

    def __getitem__(self, idx):
        """
        Allows access via index to the underlying :class:`fileseq.frameset.FrameSet`.

        :type idx: int
        :param idx: the desired index
        :rtype: int
        """
        return self.index(idx)

    def __len__(self):
        """
        The length (number of files) represented by this :class:`FileSequence`.

        :rtype: int
        """
        if not self._frameSet or not self._zfill:
            return 1
        return len(self._frameSet)

    def __str__(self):
        """
        String representation of this :class:`FileSequence`.

        :rtype: str
        """
        frameSet = str(self._frameSet or "")
        return "".join((
            self._dir,
            self._base,
            frameSet,
            self._pad if frameSet else "",
            self._ext))

    @staticmethod
    def yield_sequences_in_list(paths):
        """
        Yield the discrete sequences within paths.  This does not try to
        determine if the files actually exist on disk, it assumes you already
        know that.

        :param paths: a list of paths
        :rtype: generator
        """
        seqs = {}
        _check = DISK_RE.match

        for match in ifilter(None, imap(_check, paths)):
            dirname, basename, frame, ext = match.groups()
            if not basename and not ext:
                continue
            key = (dirname, basename, ext)
            seqs.setdefault(key, set())
            if frame:
                seqs[key].add(frame)

        for (dirname, basename, ext), frames in seqs.iteritems():
            # build the FileSequence behind the scenes, rather than dupe work
            seq = FileSequence.__new__(FileSequence)
            seq._dir = dirname or ''
            seq._base = basename or ''
            seq._ext = ext or ''
            if frames:
                seq._frameSet = FrameSet(set(imap(int, frames))) if frames else None
                seq._pad = FileSequence.getPaddingChars(min(imap(len, frames)))
            else:
                seq._frameSet = None
                seq._pad = ''
            seq.__init__(str(seq))
            yield seq

    @staticmethod
    def findSequencesInList(paths):
        """
        Returns the list of discrete sequences within paths.  This does not try
        to determine if the files actually exist on disk, it assumes you
        already know that.

        :param paths: a list of paths
        :rtype: list
        """
        return list(FileSequence.yield_sequences_in_list(paths))

    @classmethod
    def findSequencesOnDisk(cls, dirpath, include_hidden=False):
        """
        Yield the sequences found in the given directory.

        :param dirpath: directory to scan
        :type include_hidden: bool
        :param include_hidden: if true, show .hidden files as well
        :rtype: list
        """
        # reserve some functions we're going to need quick access to
        _not_hidden = lambda f: not f.startswith('.')
        _join = os.path.join

        # Get just the immediate files under the dir.
        # Avoids testing the os.listdir() for files as
        # a second step.
        ret = next(os.walk(dirpath), None)
        files = ret[-1] if ret else []

        # collapse some generators to get us the files that match our regex
        if not include_hidden:
            files = ifilter(_not_hidden, files)

        # Ensure our dirpath ends with a path separator, so
        # that we can control which sep is used during the 
        # os.path.join
        sep = utils._getPathSep(dirpath)
        if not dirpath.endswith(sep):
            dirpath += sep

        files = (_join(dirpath, f) for f in files)
        files = list(files)

        return list(FileSequence.yield_sequences_in_list(files))

    @classmethod
    def findSequenceOnDisk(cls, pattern, strictPadding=False):
        """
        Search for a specific sequence on disk.

        The padding characters used in the pattern are used to filter the 
        frame values of the files on disk (if strictPadding is True).

        :Example:
            # Find sequence matching basename and extension, and a wildcard for
            # any frame.
            # returns bar.1.exr bar.10.exr, bar.100.exr, bar.1000.exr, inclusive
            >>> findSequenceOnDisk("seq/bar@@@@.exr")

            # Find exactly 4-padded sequence, i.e. seq/bar1-100#.exr
            # returns only frames bar1000.exr through bar9999.exr
            >>> findSequenceOnDisk("seq/bar#.exr", strictPadding=True)

           
        :param pattern: the sequence pattern being searched for
        :rtype: str
        :param strictPadding: if True, ignore files with padding length different from pattern
        :rtype: bool
        :raises: :class:`fileseq.exceptions.FileSeqException` if no sequence is found on disk
        """
        seq = cls(pattern)

        if seq.frameRange() == '' and seq.padding() == '':
            if os.path.isfile(pattern):
                return seq

        patt = seq.format('{dirname}{basename}*{extension}')

        ext = seq.extension()
        basename = seq.basename()
        pad = seq.padding()

        globbed = iglob(patt)
        if pad and strictPadding:
            globbed = cls._filterByPaddingNum(globbed, seq.zfill())

        matches = cls.yield_sequences_in_list(globbed)
        for match in matches:
            if match.basename() == basename and match.extension() == ext:
                return match

        msg = 'no sequence found on disk matching {0}'
        raise FileSeqException(msg.format(pattern))

    @classmethod
    def _filterByPaddingNum(cls, iterable, num):
        """
        Yield only path elements from iterable which have a frame
        padding that matches the given target padding number
        """
        _check = DISK_RE.match

        for item in iterable:
            # Add a filter for paths that don't match the frame
            # padding of a given number
            matches = _check(item)
            if matches:
                actualNum = len(matches.group(3))
                if actualNum != num:
                    continue
            
            yield item

    @staticmethod
    def getPaddingChars(num):
        """
        Given a particular amount of padding, return the proper padding characters.
        """
        if num == 0:
            return "@"
        if num % 4 == 0:
            return "#" * (num // 4)
        else:
            return "@" * num

    @staticmethod
    def getPaddingNum(chars):
        """
        Given a supported group of padding characters, return the amount of padding.

        :type chars: str
        :param chars: a supported group of padding characters
        :rtype: int
        :raises: ValueError if unsupported padding character is detected
        """
        match = PRINTF_SYNTAX_PADDING_RE.match(chars)
        if match:
            return int(match.group(1))

        try:
            return sum([PAD_MAP[char] for char in chars])
        except KeyError:
            msg = "Detected an unsupported padding character: \"{}\"."
            msg += " Supported padding characters: {} or printf syntax padding"
            msg += " %<int>d"
            raise ValueError(msg.format(char, str(PAD_MAP.keys())))



#! /usr/bin/env python
"""
exceptions - Exception subclasses relevant to fileseq operations.
"""

class FileSeqException(ValueError):
    """
    Thrown for general exceptions handled by FileSeq.
    """
    pass

class MaxSizeException(ValueError):
    """
	Thrown when a range exceeds allowable size.
    """
    pass

class ParseException(FileSeqException):
    """
    Thrown after a frame range or file sequence parse error.
    """
    pass



#!/usr/bin/python

from __future__ import division

import unittest
import cPickle
import sys
import os
import re
import types
from itertools import chain
import string


TEST_DIR = os.path.abspath(os.path.dirname(__file__))
SRC_DIR = os.path.join(TEST_DIR, "../src")
sys.path.insert(0, SRC_DIR)
os.chdir(TEST_DIR)

from fileseq import (FrameSet, 
                     FileSequence, 
                     framesToFrameRange, 
                     findSequencesOnDisk,
                     findSequenceOnDisk, 
                     padFrameRange, 
                     getPaddingChars, 
                     getPaddingNum, 
                     ParseException,
                     FileSeqException)

from fileseq import constants, exceptions, utils
from fileseq.constants import PAD_MAP


def _getCommonPathSep(path):
    """
    Find the most common path seperator character used
    in a given path. Because windows supports both forward
    and backward sep characters, find the most consistently
    used.
    Defaults to ``os.sep``

    :type path: str
    :param path: A path to check for the most common sep
    :rtype: str
    """
    sep = os.sep
    count = 0
    for nextSep in ('/', '\\'):
        if path.count(nextSep) > count:
            sep = nextSep
    return sep

import fileseq.utils
fileseq.utils._getPathSep = _getCommonPathSep


def _yrange(first, last=None, incr=1):
    """
    Simple value generator for the 1-20y5 syntax.
    :param first: as per xrange
    :param last: as per xrange
    :param incr: as per xrange
    :return: generator
    """
    if last is None:
        first, last = 0, first
    whole = range(first, last, 1 if incr >= 0 else -1)
    filt = set(whole[::abs(incr)])
    for i in whole:
        if i not in filt:
            yield i


def _srange(first, last=None, incr=1):
    """
    Simple value generator for the 1-20:5 syntax.
    :param first: as per xrange
    :param last: as per xrange
    :param incr: as per xrange
    :return: generator
    """
    if last is None:
        first, last = 0, first
    whole = range(first, last, 1 if incr >= 0 else -1)
    sent = set()
    for stagger in xrange(abs(incr), 0, -1):
        for i in whole[::stagger]:
            if i not in sent:
                sent.add(i)
                yield i


def _uchain(*args):
    """
    As per itertools.chain, but will only yield items not previously yielded.
    :param args: one or more iterables to chain
    :return: generator
    """
    sent = set()
    for i in chain(*args):
        if i not in sent:
            yield i
            sent.add(i)


FRAME_SET_SHOULD_SUCCEED = [
    # the null value
    ("Empty", '', []),
    # individual frames
    ('Zero', '0', [0]),
    ('NegZero', '-0', [0]),
    ('Pos', '1', [1]),
    ('Neg', '-1', [-1]),
    # permutations on comma separated individual frames
    ('DupePos', '1,1,1', [1]),
    ('DupeNeg', '-1,-1,-1', [-1]),
    ('DupeMix', '-1,1,-1,1', [-1,1]),
    ('CommaSepPos', '1,3,17', [1,3,17]),
    ('CommaSepNeg', '-1,-3,-17', [-1,-3,-17]),
    ('CommaSepMix', '1,-3,17', [1,-3,17]),
    ('CommaSepPosInv', '17,3,1', [17,3,1]),
    ('CommaSepNegInv', '-17,-3,-1', [-17,-3,-1]),
    ('CommaSepMixInv', '17,-3,1', [17,-3,1]),
    ('CommaSepMixInv', '17,-3,1', [17,-3,1]),
    ("CommaTrailing", "1,", [1]),
    ("CommaLeading", ",1", [1]),
    ("CommaDupes", "1,,,,,,2,,,,,3,,,", [1,2,3]),
    # args that str(arg) cast to a valid FrameSet
    ('PosInt', 1, [1]),
    ('NegInt', -1, [-1]),
    ('FrameSet', FrameSet("1-20"), list(xrange(1,21))),
    # unicode args that are the equivalent of a valid FrameSet
    ('UnicodeEquivalentRange', u'-1--20', list(xrange(-1,-21,-1))),
    ('UnicodeEquivalentRangeChunk', u'-1--20x5', list(xrange(-1,-21,-5))),
    ('UnicodeEquivalentRangeFill', u'-1--20y5', list(_yrange(-1,-21,-5))),
    ('UnicodeEquivalentRangeStagger', u'-1--20:5', list(_srange(-1,-21,-5))),
]

LO_RANGES = [
    # low value permutations of signed integer ranges, these will all be individually tested
    ('PosToPos', '1-20', list(xrange(1,21,1))),
    ('NegToPos', '-1-20', list(xrange(-1,21,1))),
    ('NegToNeg', '-1--20', list(xrange(-1,-21,-1))),
    ('PosToNeg', '1--20', list(xrange(1,-21,-1))),
    ('PosToPosInv', '20-1', list(xrange(20,0,-1))),
    ('NegToPosInv', '-20-1', list(xrange(-20,2,1))),
    ('NegToNegInv', '-20--1', list(xrange(-20,0,1))),
    ('PosToNegInv', '20--1', list(xrange(20,-2,-1))),
    ('PosToPosChunk', '1-20x5', list(xrange(1,21,5))),
    ('NegToPosChunk', '-1-20x5', list(xrange(-1,21,5))),
    ('NegToNegChunk', '-1--20x5', list(xrange(-1,-21,-5))),
    ('PosToNegChunk', '1--20x5', list(xrange(1,-21,-5))),
    ('PosToPosChunkInv', '20-1x5', list(xrange(20,0,-5))),
    ('NegToPosChunkInv', '-20-1x5', list(xrange(-20,2,5))),
    ('NegToNegChunkInv', '-20--1x5', list(xrange(-20,0,5))),
    ('PosToNegChunkInv', '20--1x5', list(xrange(20,-2,-5))),
    ('PosToPosFill', '1-20y5', list(_yrange(1,21,5))),
    ('NegToPosFill', '-1-20y5', list(_yrange(-1,21,5))),
    ('NegToNegFill', '-1--20y5', list(_yrange(-1,-21,-5))),
    ('PosToNegFill', '1--20y5', list(_yrange(1,-21,-5))),
    ('PosToPosFillInv', '20-1y5', list(_yrange(20,0,-5))),
    ('NegToPosFillInv', '-20-1y5', list(_yrange(-20,2,5))),
    ('NegToNegFillInv', '-20--1y5', list(_yrange(-20,0,5))),
    ('PosToNegFillInv', '20--1y5', list(_yrange(20,-2,-5))),
    ('PosToPosStagger', '1-20:5', list(_srange(1,21,5))),
    ('NegToPosStagger', '-1-20:5', list(_srange(-1,21,5))),
    ('NegToNegStagger', '-1--20:5', list(_srange(-1,-21,-5))),
    ('PosToNegStagger', '1--20:5', list(_srange(1,-21,-5))),
    ('PosToPosStaggerInv', '20-1:5', list(_srange(20,0,-5))),
    ('NegToPosStaggerInv', '-20-1:5', list(_srange(-20,2,5))),
    ('NegToNegStaggerInv', '-20--1:5', list(_srange(-20,0,5))),
    ('PosToNegStaggerInv', '20--1:5', list(_srange(20,-2,-5)))]

HI_RANGES = [
    # high value permutations of signed integer ranges, these will be permuted with the LO_RANGES for testing
    ('PosToPos', '21-30', list(xrange(21,31,1))),
    ('NegToPos', '-21-30', list(xrange(-21,31,1))),
    ('NegToNeg', '-21--30', list(xrange(-21,-31,-1))),
    ('PosToNeg', '21--30', list(xrange(21,-31,-1))),
    ('PosToPosInv', '30-21', list(xrange(30,20,-1))),
    ('NegToPosInv', '-30-21', list(xrange(-30,22,1))),
    ('NegToNegInv', '-30--21', list(xrange(-30,-20,1))),
    ('PosToNegInv', '30--21', list(xrange(30,-22,-1))),
    ('PosToPosChunk', '21-30x5', list(xrange(21,31,5))),
    ('NegToPosChunk', '-21-30x5', list(xrange(-21,31,5))),
    ('NegToNegChunk', '-21--30x5', list(xrange(-21,-31,-5))),
    ('PosToNegChunk', '21--30x5', list(xrange(21,-31,-5))),
    ('PosToPosChunkInv', '30-21x5', list(xrange(30,20,-5))),
    ('NegToPosChunkInv', '-30-21x5', list(xrange(-30,22,5))),
    ('NegToNegChunkInv', '-30--21x5', list(xrange(-30,-20,5))),
    ('PosToNegChunkInv', '30--21x5', list(xrange(30,-22,-5))),
    ('PosToPosFill', '21-30y5', list(_yrange(21,31,5))),
    ('NegToPosFill', '-21-30y5', list(_yrange(-21,31,5))),
    ('NegToNegFill', '-21--30y5', list(_yrange(-21,-31,-5))),
    ('PosToNegFill', '21--30y5', list(_yrange(21,-31,-5))),
    ('PosToPosFillInv', '30-21y5', list(_yrange(30,20,-5))),
    ('NegToPosFillInv', '-30-21y5', list(_yrange(-30,22,5))),
    ('NegToNegFillInv', '-30--21y5', list(_yrange(-30,-20,5))),
    ('PosToNegFillInv', '30--21y5', list(_yrange(30,-22,-5))),
    ('PosToPosStagger', '21-30:5', list(_srange(21,31,5))),
    ('NegToPosStagger', '-21-30:5', list(_srange(-21,31,5))),
    ('NegToNegStagger', '-21--30:5', list(_srange(-21,-31,-5))),
    ('PosToNegStagger', '21--30:5', list(_srange(21,-31,-5))),
    ('PosToPosStaggerInv', '30-21:5', list(_srange(30,20,-5))),
    ('NegToPosStaggerInv', '-30-21:5', list(_srange(-30,22,5))),
    ('NegToNegStaggerInv', '-30--21:5', list(_srange(-30,-20,5))),
    ('PosToNegStaggerInv', '30--21:5', list(_srange(30,-22,-5)))]

for lo in LO_RANGES:
    FRAME_SET_SHOULD_SUCCEED.append(lo)
    for hi in HI_RANGES:
        name = 'CommaSep{0}To{1}'.format(lo[0], hi[0])
        test = ','.join([lo[1], hi[1]])
        expect = list(_uchain(lo[2], hi[2]))
        FRAME_SET_SHOULD_SUCCEED.append((name, test, expect))

FRAME_SET_SHOULD_FAIL = [
    ("PosWChunkChar", "1x5"),
    ("NegWChunkChar", "-1x5"),
    ("PosWFillChar", "1y5"),
    ("NegWFillChar", "-1y5"),
    ("PosWStaggerChar", "1:5"),
    ("NegWStaggerChar", "-1:5"),
    ("PosWSepChar", "1-"),
    ("NegWSepChar", "-1-"),
    ("BadAlphaChars", "bilbo"),
    ("RangeWDupeSpecialChar", "1-20x:y5"),
    ("RangeWBadCaseChunkChar", "1-20X5"),
    ("RangeWBadCaseFillChar", "1-20Y5"),
    ("RangeWChunkZero", "1-20x0"),
    ("RangeWFillZero", "1-20y0"),
    ("RangeWStaggerZero", "1-20:0"),
    ("RangeWNegChunk", "1-20x-5"),
    ("RangeWNegFill", "1-20y-5"),
    ("RangeWNegStagger", "1-20:-5"),
    ("FloatFrames", "1.0-10.5"),
    ("ActualNone", None),
]

class TestUtils(unittest.TestCase):
    
    def testXrange(self):
        # Test that a platform-specific xrange does not produce OverflowError
        xrng = utils.xrange(1, sys.maxint)
        self.assertTrue(len(xrng) != 0)
    
    
class TestFrameSet(unittest.TestCase):
    """
    Exercise the TestFrame object.  Due to the sheer number of permutations, we'll add most tests dynamically.
    """

    def _check___init___range(self, test, expect):
        """
        Harness to test if the FrameSet.__init__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        m = u'FrameSet("{0}")._frange != {0}: got {1}'
        r = f._frange
        self.assertEqual(r, str(test), m.format(test, r))
        m = u'FrameSet("{0}")._frange returns {1}: got {2}'
        self.assertIsInstance(r, str, m.format(test, str, type(r)))

    def _check___init___items(self, test, expect):
        """
        Harness to test if the FrameSet.__init__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        m = u'FrameSet("{0}")._items != {1}: got {2}'
        r = f._items
        self.assertEqual(r, set(expect), m.format(test, set(expect), r))
        m = u'FrameSet("{0}")._FrameSet__items returns {1}: got {2}'
        self.assertIsInstance(r, frozenset, m.format(test, frozenset, type(r)))

    def _check___init___order(self, test, expect):
        """
        Harness to test if the FrameSet.__init__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        m = u'FrameSet("{0}")._order != {1}: got {2}'
        r = f._order
        self.assertEqual(r, tuple(expect), m.format(test, tuple(expect), r))
        m = u'FrameSet("{0}")._order returns {1}: got {2}'
        self.assertIsInstance(r, tuple, m.format(test, tuple, type(r)))

    def _check___init____malformed(self, test):
        """
        Harness to test if the FrameSet.__init__ call properly handles malformed strings.
        :param test: the string to pass to FrameSet
        :return: None
        """
        try:
            r = FrameSet(test)
        except ParseException as err:
            r = err
        except Exception as err:
            r = err
        m = u'FrameSet("{0}") should fail: got {1}'
        self.assertIsInstance(r, ParseException, m.format(test, r))

    def _check___str__(self, test, expect):
        """
        Harness to test if the FrameSet.__str__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        m = u'str(FrameSet("{0}")) != {0}: got {1}'
        r = str(f)
        self.assertEqual(r, str(test), m.format(test, r))
        m = u'str(FrameSet("{0}")) returns {1}: got {2}'
        self.assertIsInstance(r, str, m.format(test, str, type(r)))

    def _check___len__(self, test, expect):
        """
        Harness to test if the FrameSet.__len__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        m = u'len(FrameSet("{0}")) != {1}: got {2}'
        r = len(f)
        self.assertEqual(r, len(expect), m.format(test, len(expect), r))
        m = u'len(FrameSet("{0}")) returns {1}: got {2}'
        self.assertIsInstance(r, int, m.format(test, int, type(r)))

    def _check___getitem__(self, test, expect):
        """
        Harness to test if the FrameSet.__getitem__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        i = len(expect) // 2
        m = u'FrameSet("{0}")[{1}] != {2}: got {3}'
        # the empty FrameSet is expected to always fail
        if not test and not expect:
            self.assertRaises(IndexError, f.__getitem__, i)
            return
        try:
            r = f[i]
        except Exception as err:
            r = repr(err)
        self.assertEqual(r, expect[i], m.format(test, i, expect[i], r))
        m = u'FrameSet("{0}")[{1}] returns {2}: got {3}'
        self.assertIsInstance(r, int, m.format(test, i, int, type(r)))
        try:
            r = f[:-1:2]
        except Exception as err:
            r = repr(err)
        e = tuple(expect[:-1:2])
        m = u'FrameSet("{0}")[:1:2] != {1}: got {2}'
        self.assertEqual(r, e, m.format(test, e, r))

    def _check_start(self, test, expect):
        """
        Harness to test if the FrameSet.start call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        m = u'FrameSet("{0}").start() != {1}: got {2}'
        # the empty FrameSet is expected to always fail
        if not test and not expect:
            self.assertRaises(IndexError, f.start)
            return
        try:
            r = f.start()
        except Exception as err:
            r = repr(err)
        self.assertEqual(r, expect[0], m.format(test, expect[0], r))
        m = u'FrameSet("{0}").start() returns {1}: got {2}'
        self.assertIsInstance(r, int, m.format(test, int, type(r)))

    def _check_end(self, test, expect):
        """
        Harness to test if the FrameSet.end call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        m = u'FrameSet("{0}").end() != {1}: got {2}'
        # the empty FrameSet is expected to always fail
        if not test and not expect:
            self.assertRaises(IndexError, f.end)
            return
        try:
            r = f.end()
        except Exception as err:
            r = repr(err)
        self.assertEqual(r, expect[-1], m.format(test, expect[-1], r))
        m = u'FrameSet("{0}").end() returns {1}: got {2}'
        self.assertIsInstance(r, int, m.format(test, int, type(r)))

    def _check_index(self, test, expect):
        """
        Harness to test if the FrameSet.index call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is expected to always fail
        if not test and not expect:
            self.assertRaises(IndexError, f.frame, 0)
            return
        i = expect[len(expect) // 2]
        m = u'FrameSet("{0}").index({1}) != {2}: got {3}'
        try:
            r = f.index(i)
        except Exception as err:
            r = repr(err)
        self.assertEqual(r, expect.index(i), m.format(test, i, expect.index(i), r))
        m = u'FrameSet("{0}").index({1}) returns {2}: got {3}'
        self.assertIsInstance(r, int, m.format(test, i, int, type(r)))

    def _check_frame(self, test, expect):
        """
        Harness to test if the FrameSet.frame call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is expected to always fail
        if not test and not expect:
            self.assertRaises(IndexError, f.frame, 0)
            return
        i = len(expect) // 2
        m = u'FrameSet("{0}").frame({1}) != {2}: got {3}'
        try:
            r = f.frame(i)
        except Exception as err:
            r = repr(err)
        self.assertEqual(r, expect[i], m.format(test, i, expect[i], r))
        m = u'FrameSet("{0}").frame({1}) returns {2}: got {3}'
        self.assertIsInstance(r, int, m.format(test, i, int, type(r)))

    def _check_hasFrameTrue(self, test, expect):
        """
        Harness to test if the FrameSet.hasFrame call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is expected to always fail
        if not test and not expect:
            self.assertFalse(f.hasFrame(1))
            return
        i = max(expect)
        m = u'FrameSet("{0}").hasFrame({1}) != {2}: got {3}'
        r = f.hasFrame(i)
        self.assertTrue(r, m.format(test, i, i in expect, r))
        m = u'FrameSet("{0}").frame({1}) returns {2}: got {3}'
        self.assertIsInstance(r, bool, m.format(test, i, bool, type(r)))

    def _check_hasFrameFalse(self, test, expect):
        """
        Harness to test if the FrameSet.hasFrame call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is expected to always fail
        if not test and not expect:
            self.assertFalse(f.hasFrame(1))
            return
        i = max(expect) + 1
        m = u'FrameSet("{0}").hasFrame({1}) != {2}: got {3}'
        r = f.hasFrame(i)
        self.assertFalse(r, m.format(test, i, i in expect, r))
        m = u'FrameSet("{0}").frame({1}) returns {2}: got {3}'
        self.assertIsInstance(r, bool, m.format(test, i, bool, type(r)))

    def _check___iter__(self, test, expect):
        """
        Harness to test if the FrameSet.__iter__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        m = u'list(FrameSet("{0}")) != {1}: got {2}'
        r = f.__iter__()
        self.assertEqual(list(r), expect, m.format(test, expect, list(r)))
        m = u'FrameSet("{0}").__iter__ returns {1}: got {2}'
        self.assertIsInstance(r, types.GeneratorType, m.format(test, types.GeneratorType, type(r)))

    def _check_canSerialize(self, test, expect):
        """
        Harness to test if the FrameSet.__getstate__ and FrameSet.__setstate__ calls allowing pickling.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        f2 = cPickle.loads(cPickle.dumps(f))
        m = u'FrameSet("{0}") does not pickle correctly'
        self.assertIsInstance(f2, FrameSet, m.format(test))
        self.assertTrue(str(f) == str(f2) and list(f) == list(f2), m.format(test))
        # test old objects being unpickled through new lib
        state = {'__frange': f._frange, '__set': set(f._items), '__list': list(f._order)}
        f2 = FrameSet.__new__(FrameSet)
        f2.__setstate__(state)
        self.assertTrue(str(f) == str(f2) and list(f) == list(f2), m.format(test))

    def _check_frameRange(self, test, expect):
        """
        Harness to test if the FrameSet.frameRange call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet always has a frameRange of ''
        if not test and not expect:
            self.assertEqual(f.frameRange(), '')
            return
        m = u'FrameSet("{0}").frameRange({1}) != "{2}": got "{3}"'
        p = '((?<![xy:])\d+)'
        l = max([max([len(i) for i in re.findall(p, str(f))]) + 1, 4])
        expect = re.sub(p, lambda m: m.group(0).rjust(l, '0'), str(f))
        try:
            r = f.frameRange(l)
        except Exception as err:
            r = repr(err)
        self.assertEqual(r, expect, m.format(test, l, expect, r))
        m = u'FrameSet("{0}").frameRange({1}) returns {2}: got {3}'
        self.assertIsInstance(r, str, m.format(test, i, str, type(r)))

    def _check_invertedFrameRange(self, test, expect):
        """
        Harness to test if the FrameSet.invertedFrameRange call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        m = u'FrameSet("{0}").invertedFrameRange() != "{1}": got "{2}"'
        r = f.invertedFrameRange()
        t = sorted(f)
        c = sorted(FrameSet(r) if r else [])
        # the empty FrameSet will always return '' for inverted and normal
        # FrameRange
        if not test and not expect:
            self.assertEqual(r, '')
        else:
            e = [i for i in xrange(t[0], t[-1]) if i not in t]
            self.assertEqual(c, e, m.format(test, e, c))
        m = u'FrameSet("{0}").invertedFrameRange() returns {1}: got {2}'
        self.assertIsInstance(r, str, m.format(test, str, type(r)))

    def _check_normalize(self, test, expect):
        """
        Harness to test if the FrameSet.normalize call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        m = u'set(FrameSet("{0}").normalize()) != {1}: got {2}'
        r = f.normalize()
        self.assertEqual(set(f), set(r), m.format(test, set(expect), set(r)))
        m = u'FrameSet("{0}").normalize() returns {1}: got {2}'
        self.assertIsInstance(r, FrameSet, m.format(test, FrameSet, type(r)))

    def _check_isFrameRange(self, test, expect):
        """
        Harness to test if the FrameSet.isFrameRange call works properly.
        :param test: the string to pass to FrameSet.isFrameRange
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        r = FrameSet.isFrameRange(test)
        m = u'FrameSet.isFrameRange("{0}") != {1}: got {2}'
        self.assertEqual(r, expect, m.format(test, expect, r))
        m = u'FrameSet.isFrameRange("{0}") returns {1}: got {2}'
        self.assertIsInstance(r, bool, m.format(test, bool, type(r)))

    def _check_fromIterable(self, expect, iterable):
        """
        Harness to test if the FrameSet.fromIterable call works properly.
        :param expect: the string to use to build the expected FrameRange, which will be normalized for comparison
        :param iterable: the iterable to test
        :return: None
        """
        e = FrameSet(expect)
        r = FrameSet.from_iterable(iterable)
        m = u'FrameSet.fromIterable({0}) != {1!r}: got {2!r}'
        self.assertEqual(r, e, m.format(iterable, e, r))
        m = u'FrameSet.fromIterable({0}) returns {1}: got {2}'
        self.assertIsInstance(r, FrameSet, m.format(expect, FrameSet, type(r)))

    def _check___repr__(self, test, expect):
        """
        Harness to test if the FrameSet.__repr__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        e = 'FrameSet("{0}")'.format(test)
        m = u'repr(FrameSet("{0}")) != {1}: got {2}'
        self.assertEqual(repr(f), e, m.format(test, e, repr(f)))
        m = u'repr(FrameSet("{0}")) returns {1}: got {2}'
        self.assertIsInstance(repr(f), str, m.format(test, str, type(repr(f))))

    def _check___reversed__(self, test, expect):
        """
        Harness to test if the FrameSet.__reversed__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        e = list(reversed(expect))
        r = reversed(f)
        m = u'reversed(FrameSet("{0}")) != {1}: got {2}'
        self.assertEqual(list(r), e, m.format(test, e, r))
        m = u'reversed(FrameSet("{0}")) returns {1}: got {2}'
        self.assertIsInstance(r, types.GeneratorType, m.format(test, types.GeneratorType, type(r)))

    def _check___contains__(self, test, expect):
        """
        Harness to test if the FrameSet.__contains__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        e = expect[-1] if len(expect) else None
        should_succeed = e in f
        e = (max(expect) + 1) if len(expect) else None
        should_fail = e in f
        m = u'{0} in FrameSet("{1}"))'
        # the empty FrameSet contains nothing
        if not test and not expect:
            self.assertFalse(should_succeed, m.format(e, test))
            self.assertFalse(should_fail, m.format(e, test))
        else:
            self.assertTrue(should_succeed, m.format(e, test))
            self.assertFalse(should_fail, m.format(e, test))
        m = u'FrameSet("{0}").__contains__ returns {1}: got {2}'
        self.assertIsInstance(should_succeed, bool, m.format(test, bool, type(should_succeed)))
        self.assertIsInstance(should_fail, bool, m.format(test, bool, type(should_fail)))

    def _check___hash__(self, test, expect):
        """
        Harness to test if the FrameSet.__hash__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        try:
            r = hash(f)
        except Exception as err:
            r = err
        m = u'hash(FrameSet("{0}")) returns {1}: got {2}'
        self.assertIsInstance(r, int, m.format(test, int, type(r)))

    def _check___lt__(self, test, expect):
        """
        Harness to test if the FrameSet.__lt__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is less than everything, except for itself
        if not test and not expect:
            self.assertTrue(f < FrameSet('1'))
            self.assertTrue(f < FrameSet('-1'))
            self.assertFalse(f < expect)
            return
        r = FrameSet.from_iterable(expect + [max(expect) + 1])
        should_succeed = f < r
        should_fail = r < f
        m = u'FrameSet("{0}") < FrameSet("{1}")'
        self.assertTrue(should_succeed, m.format(test, r))
        self.assertFalse(should_fail, m.format(r, test))
        m = u'FrameSet("{0}") < FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(should_succeed, bool, m.format(test, r, bool, type(should_succeed)))
        self.assertIsInstance(should_fail, bool, m.format(r, test, bool, type(should_fail)))

    def _check___le__(self, test, expect):
        """
        Harness to test if the FrameSet.__le__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is less than everything, equal only to itself
        if not test and not expect:
            self.assertTrue(f <= FrameSet('1'))
            self.assertTrue(f <= FrameSet('-1'))
            self.assertTrue(f <= expect)
            return
        for i in [expect, expect + [max(expect) + 1]]:
            r = FrameSet.from_iterable(i)
            should_succeed = f <= r
            m = u'FrameSet("{0}") <= FrameSet("{1}")'
            self.assertTrue(should_succeed, m.format(test, r))
            m = u'FrameSet("{0}") <= FrameSet("{1}") returns {2}: got {3}'
            self.assertIsInstance(should_succeed, bool, m.format(test, r, bool, type(should_succeed)))

    def _check___eq__(self, test, expect):
        """
        Harness to test if the FrameSet.__eq__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        r = FrameSet(','.join((str(i) for i in expect)))
        should_succeed = f == r
        m = u'FrameSet("{0}") == FrameSet("{1}")'
        self.assertTrue(should_succeed, m.format(test, r))
        m = u'FrameSet("{0}") == FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(should_succeed, bool, m.format(test, r, bool, type(should_succeed)))

    def _check___ne__(self, test, expect):
        """
        Harness to test if the FrameSet.__ne__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is not equal to anything, except for itself
        if not test and not expect:
            self.assertTrue(f != FrameSet('1'))
            self.assertTrue(f != FrameSet('-1'))
            self.assertFalse(f != expect)
            return
        r = FrameSet(','.join((str(i) for i in (expect + [max(expect) + 1]))))
        should_succeed = f != r
        m = u'FrameSet("{0}") != FrameSet("{1}")'
        self.assertTrue(should_succeed, m.format(test, r))
        m = u'FrameSet("{0}") != FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(should_succeed, bool, m.format(test, r, bool, type(should_succeed)))

    def _check___ge__(self, test, expect):
        """
        Harness to test if the FrameSet.__ge__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is greater than nothing, except for itself
        if not test and not expect:
            self.assertFalse(f >= FrameSet('1'))
            self.assertFalse(f >= FrameSet('-1'))
            self.assertTrue(f >= expect)
            return
        for i in [expect, expect[:-1]]:
            try:
                r = FrameSet.from_iterable(i)
            except ParseException:
                # this will happen if len(expect) == 1
                continue
            should_succeed = f >= r
            m = u'FrameSet("{0}") >= FrameSet("{1}"'
            self.assertTrue(should_succeed, m.format(test, r))
            m = u'FrameSet("{0}") >= FrameSet("{1}") returns {2}: got {3}'
            self.assertIsInstance(should_succeed, bool, m.format(test, r, bool, type(should_succeed)))

    def _check___gt__(self, test, expect):
        """
        Harness to test if the FrameSet.__gt__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is greater than nothing, except for itself
        if not test and not expect:
            self.assertFalse(f > FrameSet('1'))
            self.assertFalse(f > FrameSet('-1'))
            self.assertFalse(f > expect)
            return
        try:
            r = FrameSet.from_iterable(expect[:-1])
        except ParseException:
            # this will happen if len(expect) == 1
            return
        should_succeed = f > r
        should_fail = r > f
        m = u'FrameSet("{0}") > FrameSet("{1}")'
        self.assertTrue(should_succeed, m.format(test, r))
        self.assertFalse(should_fail, m.format(r, test))
        m = u'FrameSet("{0}") > FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(should_succeed, bool, m.format(test, r, bool, type(should_succeed)))
        self.assertIsInstance(should_fail, bool, m.format(r, test, bool, type(should_fail)))

    def _check___and__(self, test, expect):
        """
        Harness to test if the FrameSet.__and__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        v = [i + max(expect) + 1 for i in expect] or range(999, 1999)
        t = FrameSet.from_iterable(v)
        r = f & t
        e = FrameSet.from_iterable(set(expect) & set(v), sort=True)
        m = u'FrameSet("{0}") & FrameSet("{1}") != FrameSet("{2}")'
        self.assertEqual(r, e, m.format(f, t, e))
        m = u'FrameSet("{0}") & FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(r, FrameSet, m.format(test, t, FrameSet, type(r)))

    def _check___rand__(self, test, expect):
        """
        Harness to test if the FrameSet.__rand__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        v = [i + max(expect) + 1 for i in expect] or range(999, 1999)
        t = FrameSet.from_iterable(v)
        r = t & f
        e = FrameSet.from_iterable(set(v) & set(expect), sort=True)
        m = u'FrameSet("{0}") & FrameSet("{1}") != FrameSet("{2}")'
        self.assertEqual(r, e, m.format(t, f, e))
        m = u'FrameSet("{0}") & FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(r, FrameSet, m.format(t, test, FrameSet, type(r)))

    def _check___sub__(self, test, expect):
        """
        Harness to test if the FrameSet.__sub__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        v = [i + max(expect) + 1 for i in expect] or range(999, 1999)
        t = FrameSet.from_iterable(v)
        r = f - t
        e = FrameSet.from_iterable(set(expect) - set(v), sort=True)
        m = u'FrameSet("{0}") - FrameSet("{1}") != FrameSet("{2}")'
        self.assertEqual(r, e, m.format(f, t, e))
        m = u'FrameSet("{0}") - FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(r, FrameSet, m.format(test, t, FrameSet, type(r)))

    def _check___rsub__(self, test, expect):
        """
        Harness to test if the FrameSet.__rsub__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        v = [i + max(expect) + 1 for i in expect] or range(999, 1999)
        t = FrameSet.from_iterable(v)
        r = t - f
        e = FrameSet.from_iterable(set(v) - set(expect), sort=True)
        m = u'FrameSet("{0}") - FrameSet("{1}") != FrameSet("{2}")'
        self.assertEqual(r, e, m.format(t, f, e))
        m = u'FrameSet("{0}") - FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(r, FrameSet, m.format(t, test, FrameSet, type(r)))

    def _check___or__(self, test, expect):
        """
        Harness to test if the FrameSet.__or__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        v = [i + max(expect) + 1 for i in expect] or range(999, 1999)
        t = FrameSet.from_iterable(v)
        r = f | t
        e = FrameSet.from_iterable(set(expect) | set(v), sort=True)
        m = u'FrameSet("{0}") | FrameSet("{1}") != FrameSet("{2}")'
        self.assertEqual(r, e, m.format(f, t, e))
        m = u'FrameSet("{0}") | FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(r, FrameSet, m.format(test, t, FrameSet, type(r)))

    def _check___ror__(self, test, expect):
        """
        Harness to test if the FrameSet.__ror__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        v = [i + max(expect) + 1 for i in expect] or range(999, 1999)
        t = FrameSet.from_iterable(v)
        r = t | f
        e = FrameSet.from_iterable(set(v) | set(expect), sort=True)
        m = u'FrameSet("{0}") | FrameSet("{1}") != FrameSet("{2}")'
        self.assertEqual(r, e, m.format(t, f, e))
        m = u'FrameSet("{0}") | FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(r, FrameSet, m.format(t, test, FrameSet, type(r)))

    def _check___xor__(self, test, expect):
        """
        Harness to test if the FrameSet.__xor__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        v = [i + max(expect) + 1 for i in expect] or range(999, 1999)
        t = FrameSet.from_iterable(v)
        r = f ^ t
        e = FrameSet.from_iterable(set(expect) ^ set(v), sort=True)
        m = u'FrameSet("{0}") ^ FrameSet("{1}") != FrameSet("{2}")'
        self.assertEqual(r, e, m.format(f, t, e))
        m = u'FrameSet("{0}") ^ FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(r, FrameSet, m.format(test, t, FrameSet, type(r)))

    def _check___rxor__(self, test, expect):
        """
        Harness to test if the FrameSet.__rxor__ call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        v = [i + max(expect) + 1 for i in expect] or range(999, 1999)
        t = FrameSet.from_iterable(v)
        r = t ^ f
        e = FrameSet.from_iterable(set(v) ^ set(expect), sort=True)
        m = u'FrameSet("{0}") ^ FrameSet("{1}") != FrameSet("{2}")'
        self.assertEqual(r, e, m.format(t, f, e))
        m = u'FrameSet("{0}") ^ FrameSet("{1}") returns {2}: got {3}'
        self.assertIsInstance(r, FrameSet, m.format(t, test, FrameSet, type(r)))

    def _check_isdisjoint(self, test, expect):
        """
        Harness to test if the FrameSet.isdisjoint call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is the disjoint of everything, including itself
        if not test and not expect:
            self.assertTrue(f.isdisjoint(FrameSet('1')))
            self.assertTrue(f.isdisjoint(FrameSet('-1')))
            self.assertTrue(f.isdisjoint(expect))
            return
        for v in [[expect[0]], expect, expect + [max(expect)+1], [i + max(expect) + 1 for i in expect]]:
            t = FrameSet.from_iterable(v)
            r = f.isdisjoint(t)
            e = set(expect).isdisjoint(v)
            m = u'FrameSet("{0}").isdisjoint(FrameSet("{1}")) != {2}'
            self.assertEquals(r, e, m.format(t, f, e))
            m = u'FrameSet("{0}").isdisjoint(FrameSet("{1}")) returns {2}: got {3}'
            self.assertIsInstance(r, bool, m.format(test, t, bool, type(r)))

    def _check_issubset(self, test, expect):
        """
        Harness to test if the FrameSet.issubset call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is the subset of everything, including itself
        if not test and not expect:
            self.assertTrue(f.issubset(FrameSet('1')))
            self.assertTrue(f.issubset(FrameSet('-1')))
            self.assertTrue(f.issubset(expect))
            return
        for v in [[expect[0]], expect, expect + [max(expect)+1], [i + max(expect) + 1 for i in expect]]:
            t = FrameSet.from_iterable(v)
            r = f.issubset(t)
            e = set(expect).issubset(v)
            m = u'FrameSet("{0}").issubset(FrameSet("{1}")) != {2}'
            self.assertEquals(r, e, m.format(t, f, e))
            m = u'FrameSet("{0}").issubset(FrameSet("{1}")) returns {2}: got {3}'
            self.assertIsInstance(r, bool, m.format(test, t, bool, type(r)))

    def _check_issuperset(self, test, expect):
        """
        Harness to test if the FrameSet.issuperset call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the empty FrameSet is the superset of everything, except itself
        if not test and not expect:
            self.assertFalse(f.issuperset(FrameSet('1')))
            self.assertFalse(f.issuperset(FrameSet('-1')))
            self.assertTrue(f.issuperset(expect))
            return
        for v in [[expect[0]], expect, expect + [max(expect)+1], [i + max(expect) + 1 for i in expect]]:
            t = FrameSet.from_iterable(v)
            r = f.issuperset(t)
            e = set(expect).issuperset(v)
            m = u'FrameSet("{0}").issuperset(FrameSet("{1}")) != {2}'
            self.assertEquals(r, e, m.format(t, f, e))
            m = u'FrameSet("{0}").issuperset(FrameSet("{1}")) returns {2}: got {3}'
            self.assertIsInstance(r, bool, m.format(test, t, bool, type(r)))

    def _check_union(self, test, expect):
        """
        Harness to test if the FrameSet.union call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the union of the empty FrameSet with any other is always the other
        if not test and not expect:
            self.assertEqual(f.union(FrameSet('1')), FrameSet('1'))
            self.assertEqual(f.union(FrameSet('-1')), FrameSet('-1'))
            self.assertEqual(f.union(expect), FrameSet.from_iterable(expect, sort=True))
            return
        for v in [[expect[0]], expect, expect + [max(expect)+1], [i + max(expect) + 1 for i in expect]]:
            t = FrameSet.from_iterable(v)
            r = f.union(t)
            e = FrameSet.from_iterable(set(expect).union(v), sort=True)
            m = u'FrameSet("{0}").union(FrameSet("{1}")) != {2}'
            self.assertEquals(r, e, m.format(t, f, e))
            m = u'FrameSet("{0}").union(FrameSet("{1}")) returns {2}: got {3}'
            self.assertIsInstance(r, FrameSet, m.format(test, t, FrameSet, type(r)))

    def _check_intersection(self, test, expect):
        """
        Harness to test if the FrameSet.intersection call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the intersection of the empty FrameSet with any other is always the empty FrameSet
        if not test and not expect:
            self.assertEqual(f.intersection(FrameSet('1')), f)
            self.assertEqual(f.intersection(FrameSet('-1')), f)
            self.assertEqual(f.intersection(expect), f)
            return
        for v in [[expect[0]], expect, expect + [max(expect)+1], [i + max(expect) + 1 for i in expect]]:
            t = FrameSet.from_iterable(v)
            r = f.intersection(t)
            e = FrameSet.from_iterable(set(expect).intersection(v), sort=True)
            m = u'FrameSet("{0}").intersection(FrameSet("{1}")) != {2}'
            self.assertEquals(r, e, m.format(t, f, e))
            m = u'FrameSet("{0}").intersection(FrameSet("{1}")) returns {2}: got {3}'
            self.assertIsInstance(r, FrameSet, m.format(test, t, FrameSet, type(r)))

    def _check_difference(self, test, expect):
        """
        Harness to test if the FrameSet.difference call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the difference of the empty FrameSet with any other is always the empty FrameSet
        if not test and not expect:
            self.assertEqual(f.intersection(FrameSet('1')), f)
            self.assertEqual(f.intersection(FrameSet('-1')), f)
            self.assertEqual(f.intersection(expect), f)
            return
        for v in [[expect[0]], expect, expect + [max(expect)+1], [i + max(expect) + 1 for i in expect]]:
            t = FrameSet.from_iterable(v)
            r = f.difference(t)
            e = FrameSet.from_iterable(set(expect).difference(v), sort=True)
            m = u'FrameSet("{0}").difference(FrameSet("{1}")) != {2}'
            self.assertEquals(r, e, m.format(t, f, e))
            m = u'FrameSet("{0}").difference(FrameSet("{1}")) returns {2}: got {3}'
            self.assertIsInstance(r, FrameSet, m.format(test, t, FrameSet, type(r)))

    def _check_symmetric_difference(self, test, expect):
        """
        Harness to test if the FrameSet.symmetric_difference call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        # the symmetric_difference of the empty FrameSet with any other is always the empty FrameSet
        if not test and not expect:
            self.assertEqual(f.intersection(FrameSet('1')), f)
            self.assertEqual(f.intersection(FrameSet('-1')), f)
            self.assertEqual(f.intersection(expect), f)
            return
        for v in [[expect[0]], expect, expect + [max(expect)+1], [i + max(expect) + 1 for i in expect]]:
            t = FrameSet.from_iterable(v)
            r = f.symmetric_difference(t)
            e = FrameSet.from_iterable(set(expect).symmetric_difference(v), sort=True)
            m = u'FrameSet("{0}").symmetric_difference(FrameSet("{1}")) != {2}'
            self.assertEquals(r, e, m.format(t, f, e))
            m = u'FrameSet("{0}").symmetric_difference(FrameSet("{1}")) returns {2}: got {3}'
            self.assertIsInstance(r, FrameSet, m.format(test, t,
                                                                FrameSet, type(r)))

    def _check_copy(self, test, expect):
        """
        Harness to test if the FrameSet.copy call works properly.
        :param test: the string to pass to FrameSet
        :param expect: the expected list of values that FrameSet will hold
        :return: None
        """
        f = FrameSet(test)
        r = f.copy()
        self.assertIsNot(f, r)
        self.assertEqual(f, r)

    def testFloatFrameValues(self):
        table = [
            ([1, 5.8, 10], [1, 5, 10]),
            ([1.5, 5, 10.2], [1, 5, 10]),
            ([1.001, 5, 10.999], [1, 5, 10]),
        ]

        for src, expected in table:
            f = FrameSet(src)
            actual = list(f)
            self.assertEqual(actual, expected)

    def testMaxFrameSize(self):
        _maxSize = constants.MAX_FRAME_SIZE
        try:
            maxSize = constants.MAX_FRAME_SIZE = 500

            # Within range
            utils.xfrange(1, 100, 1, maxSize=-1)
            utils.xfrange(1, 100, 1, maxSize=100)
            FrameSet('1-%d' % maxSize)

            # Shoult not be allowed
            self.assertRaises(exceptions.MaxSizeException, utils.xfrange, 1, 100, 1, maxSize=50)
            self.assertRaises(exceptions.MaxSizeException, FrameSet, '1-%d' % (maxSize+1))

            # Inverting would produce a huge new range
            fs = FrameSet('1,%d' % (maxSize+3))
            self.assertRaises(exceptions.MaxSizeException, fs.invertedFrameRange)

        finally:
            constants.MAX_FRAME_SIZE = _maxSize


# due to the sheer number of combinations, we build the bulk of our tests on to TestFrameSet dynamically
for name, tst, exp in FRAME_SET_SHOULD_SUCCEED:
    setattr(
        TestFrameSet, 'testFrameSet%sInitSetsRange' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___init___range(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sInitSetsItems' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___init___items(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sInitSetsOrder' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___init___order(self, t, e))
    setattr(
        TestFrameSet, 'testFromIterable%s' % name,
        lambda self, e=tst, i=exp: TestFrameSet._check_fromIterable(self, e, i))
    setattr(
        TestFrameSet, 'testFrameSet%sIndex' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_index(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sFrame' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_frame(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sHasFrameTrue' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_hasFrameTrue(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sHasFrameFalse' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_hasFrameTrue(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sStart' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_start(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sEnd' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_end(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sFrameRange' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_frameRange(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sInvertedFrameRange' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_invertedFrameRange(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sNormalize' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_normalize(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sSerialize' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_canSerialize(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sGetItem' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___getitem__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sLen' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___len__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sStr' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___str__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sRepr' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___repr__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sIter' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___iter__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sReversed' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___reversed__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sContains' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___contains__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sHash' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___hash__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sLessThan' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___lt__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sLessEqual' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___le__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sEqual' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___eq__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sNotEqual' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___ne__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sGreaterEqual' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___ge__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sGreaterThan' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___gt__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sAnd' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___and__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sRightAnd' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___rand__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sSub' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___sub__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sRightSub' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___rsub__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sOr' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___or__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sRightOr' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___ror__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sExclusiveOr' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___xor__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sRightExclusiveOr' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check___rxor__(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sIsDisjoint' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_isdisjoint(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sIsSubset' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_issubset(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sIsSubset' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_issuperset(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sUnion' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_union(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sIntersection' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_intersection(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sDifference' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_difference(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sSymmetricDifference' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_symmetric_difference(self, t, e))
    setattr(
        TestFrameSet, 'testFrameSet%sCopy' % name,
        lambda self, t=tst, e=exp: TestFrameSet._check_copy(self, t, e))
    setattr(
        TestFrameSet, 'testIsFrameRange%sShouldSucceed' % name,
        lambda self, t=tst: TestFrameSet._check_isFrameRange(self, t, True))

for name, tst in FRAME_SET_SHOULD_FAIL:
    setattr(
        TestFrameSet, 'testFrameSet%sInitHandlesMalformed' % name,
        lambda self, t=tst: TestFrameSet._check___init____malformed(self, t))
    setattr(
        TestFrameSet, 'testIsFrameRange%sShouldFail' % name,
        lambda self, t=tst: TestFrameSet._check_isFrameRange(self, t, False))


class TestBase(unittest.TestCase):

    RX_PATHSEP = re.compile(r'[/\\]')

    def assertEquals(self, a, b):
        # Make sure string paths are compared with normalized
        # path separators
        if isinstance(a, basestring) and isinstance(b, basestring):
            if self.RX_PATHSEP.search(a) and self.RX_PATHSEP.search(b): 
                a = os.path.normpath(a)
                b = os.path.normpath(b)

        super(TestBase, self).assertEquals(a, b)

    def assertEqual(self, a, b):
        self.assertEquals(a, b)
    
    def assertEqualPaths(self, a, b):
        return super(TestBase, self).assertEquals(self.toNormpaths(a), self.toNormpaths(b))
    
    def toNormpaths(self, collection):
        if isinstance(collection, basestring):
            return os.path.normpath(collection)
        return sorted(map(os.path.normpath, collection))
    

class TestFramesToFrameRange(unittest.TestCase):
    """
    Exercise the frameToRange func.  Due to the sheer number of permutations, we'll add most tests dynamically.
    """

    def _check_frameToRangeEquivalence(self, test, expect):
        f = FrameSet(test)
        r = FrameSet(framesToFrameRange(expect, sort=False))
        m = '{0!r} != {1!r}'
        self.assertEqual(f, r, m.format(f, r))

# due to the sheer number of combinations, we build the bulk of our tests on to TestFramesToFrameRange dynamically
for name, tst, exp in FRAME_SET_SHOULD_SUCCEED:
    setattr(
        TestFramesToFrameRange, 'testFramesToRangeEquivalence%s' % name,
        lambda self, t=tst, e=exp: TestFramesToFrameRange._check_frameToRangeEquivalence(self, t, e))


class TestFileSequence(TestBase):

    def testSeqGettersType1(self):
        seq = FileSequence("/foo/boo.1-5#.exr")
        self.assertEquals(5, len(seq))
        self.assertEquals("/foo/", seq.dirname())
        self.assertEquals("boo.", seq.basename())
        self.assertEquals("#", seq.padding())
        self.assertEquals(".exr", seq.extension())

        self.assertEquals("/foo/boo.9999.exr", seq.frame(9999))
        self.assertEquals("/foo/boo.0001.exr", seq[0])
        self.assertEquals("/foo/boo.0001.exr", seq.index(0))

    def testSeqGettersType2(self):
        seq = FileSequence("/foo/boo1-5#.exr")
        self.assertEquals(5, len(seq))
        self.assertEquals("/foo/", seq.dirname())
        self.assertEquals("boo", seq.basename())
        self.assertEquals("#", seq.padding())
        self.assertEquals(".exr", seq.extension())

        self.assertEquals("/foo/boo9999.exr", seq.frame(9999))
        self.assertEquals("/foo/boo0001.exr", seq[0])
        self.assertEquals("/foo/boo0001.exr", seq.index(0))

    def testSeqGettersPrintf(self):
        seq = FileSequence("/foo/boo.1-5%04d.exr")
        self.assertEquals(5, len(seq))
        self.assertEquals("/foo/", seq.dirname())
        self.assertEquals("boo.", seq.basename())
        self.assertEquals("%04d", seq.padding())
        self.assertEquals(".exr", seq.extension())

        self.assertEquals("/foo/boo.9999.exr", seq.frame(9999))
        self.assertEquals("/foo/boo.0001.exr", seq[0])
        self.assertEquals("/foo/boo.0001.exr", seq.index(0))

    def testSetDirname(self):
        seq = FileSequence("/foo/bong.1-5@.exr")
        seq.setDirname("/bing/")
        self.assertEquals("/bing/bong.1.exr", seq[0])

    def testSetBasename(self):
        seq = FileSequence("/foo/bong.1-5@.exr")
        seq.setBasename("bar.")
        self.assertEquals("/foo/bar.1.exr", seq[0])

    def testSetPadding(self):
        seq = FileSequence("/foo/bong.1-5@.exr")
        seq.setPadding("#")
        self.assertEquals("/foo/bong.0001.exr", seq[0])

        seq.setPadding("%02d")
        self.assertEquals("/foo/bong.01.exr", seq[0])

    def testSetFrameSet(self):
        seq = FileSequence("/cheech/chong.1-5#.exr")
        seq.setFrameSet(FrameSet("10-20"))
        self.assertEquals("/cheech/chong.10-20#.exr", str(seq))

    def testSetFrameRange(self):
        seq = FileSequence("/cheech/chong.1-5#.exr")
        seq.setFrameRange("10-20")
        self.assertEquals("/cheech/chong.10-20#.exr", str(seq))

    def testFrame(self):
        seq = FileSequence("/foo/bar/bing.#.exr")
        self.assertEquals("/foo/bar/bing.0001.exr", seq.frame(1))
        self.assertEquals("/foo/bar/bing.#.exr", seq.frame("#"))

        seq = FileSequence("/foo/bar/bing.%04d.exr")
        self.assertEquals("/foo/bar/bing.0001.exr", seq.frame(1))
        self.assertEquals("/foo/bar/bing.%04d.exr", seq.frame("%04d"))

    def testIter(self):
        known = set ([
            "/cheech/chong.0001.exr",
            "/cheech/chong.0003.exr",
            "/cheech/chong.0005.exr"
        ])
        seq = FileSequence("/cheech/chong.1,3,5#.exr")
        self.assertFalse(known.difference(seq))

    def testFormat(self):
        seq = FileSequence("/cheech/chong.1-10,30,40#.exr")
        self.assertEquals("chong.0001-0010,0030,0040#.exr", str(seq.format()))
        self.assertEquals("0011-0029,0031-0039", seq.format("{inverted}"))

        seq = findSequencesOnDisk("broken_seq")[0]
        self.assertEquals("0000-0002,0004,0006-0008", seq.format("{range}"))
        self.assertEquals("broke.0000-0002,0004,0006-0008#.exr", seq.format())
        seq = findSequencesOnDisk("step_seq")[0]
        self.assertEquals("step_seq/step1.1-13x4,14-17#.exr", str(seq))
    
    def testFormatInverted(self):
        _maxSize = constants.MAX_FRAME_SIZE
        try:
            maxSize = constants.MAX_FRAME_SIZE = 500

            # Test catching error for large inverted range
            seq = FileSequence("/path/to/file.1,%d#.ext" % (constants.MAX_FRAME_SIZE+3))
            self.assertRaises(exceptions.MaxSizeException, seq.format, '{inverted}')

        finally:
            constants.MAX_FRAME_SIZE = _maxSize

    def testSplit(self):
        seqs = FileSequence("/cheech/chong.1-10,30,40#.exr").split()
        self.assertEquals("/cheech/chong.0001-0010#.exr", str(seqs[0]))
        self.assertEquals("/cheech/chong.0030#.exr", str(seqs[1]))
        self.assertEquals("/cheech/chong.0040#.exr", str(seqs[2]))

    def testMissingPeriods(self):
        seqs = FileSequence("/path/to/something_1-10#_exr")
        self.assertEquals("/path/to/something_0001_exr", seqs.index(0))

        seqs = FileSequence("/path/to/something_1-10%04d_exr")
        self.assertEquals("/path/to/something_0001_exr", seqs.index(0))

    def testNumericFilename(self):
        seqs = FileSequence("/path/to/1-10#.exr")
        self.assertEquals("/path/to/0001.exr", seqs.index(0))

        seqs = FileSequence("/path/to/1-10%04d.exr")
        self.assertEquals("/path/to/0001.exr", seqs.index(0))

    def testNoPlaceholder(self):
        expected = "/path/to/file.mov"
        seqs = FileSequence(expected)

        self.assertEquals(expected, seqs.index(0))
        self.assertEquals(expected, seqs.frame(0))
        self.assertEquals(expected, seqs[0])
        self.assertEquals(None, seqs.frameSet())
        self.assertEquals("", seqs.frameRange())
        self.assertEquals("", seqs.invertedFrameRange())
        self.assertEquals(1, len(seqs))

        seqs.setFrameRange("1-100")

        for i in xrange(0,100):
            self.assertEquals(expected, seqs.index(i))
            self.assertEquals(expected, seqs.frame(i+1))
            self.assertEquals(expected, seqs[i])
        self.assertEquals(1, len(seqs))

        seqs.setPadding("#")
        self.assertEquals(100, len(seqs))

    def testNoPlaceholderNumExt(self):
        basename = 'file'
        exts = ('.7zip', '.mp4')

        for ext in exts:
            expected = basename + ext
            seqs = FileSequence(expected)

            self.assertEquals(ext, seqs.extension())
            self.assertEquals(basename, seqs.basename())
            self.assertEquals(expected, str(seqs))

    def testSplitXY(self):
        seqs = FileSequence("/cheech/0-9x1/chong.1-10#.exr")
        self.assertEquals("/cheech/0-9x1/chong.0001.exr", seqs.index(0))

    def testSerialization(self):
        fs = FileSequence("/path/to/file.1-100x2#.exr")
        s = cPickle.dumps(fs, cPickle.HIGHEST_PROTOCOL)
        fs2 = cPickle.loads(s)
        self.assertEquals(str(fs), str(fs2))
        self.assertEquals(len(fs), len(fs2))

        fs = FileSequence("/path/to/file.1-100x2%04d.exr")
        s = cPickle.dumps(fs, cPickle.HIGHEST_PROTOCOL)
        fs2 = cPickle.loads(s)
        self.assertEquals(str(fs), str(fs2))
        self.assertEquals(len(fs), len(fs2))

    def testHasVersionNoFrame(self):
        fs = FileSequence("/path/to/file_v2.exr")
        self.assertEquals(fs.start(), 0)
        self.assertEquals(fs.end(), 0)
        self.assertEquals(fs.padding(), '')
        self.assertEquals(fs.extension(), '.exr')
        self.assertEquals(str(fs), "/path/to/file_v2.exr")

    def testHasFrameNoVersion(self):
        fs = FileSequence("/path/to/file.2.exr")
        self.assertEquals(fs.start(), 2)
        self.assertEquals(fs.end(), 2)
        self.assertEquals(fs.padding(), '@')
        self.assertEquals(fs.extension(), '.exr')
        self.assertEquals(str(fs), "/path/to/file.2@.exr")

    def testNoFrameNoVersionNoExt(self):
        fs = FileSequence("/path/to/file")
        self.assertEquals(fs.start(), 0)
        self.assertEquals(fs.end(), 0)
        self.assertEquals(fs.padding(), '')
        self.assertEquals(fs.dirname(), '/path/to/')
        self.assertEquals(fs.basename(), 'file')
        self.assertEquals(fs.extension(), '')
        self.assertEquals(str(fs), "/path/to/file")

        fs = FileSequence("file")
        self.assertEquals(fs.start(), 0)
        self.assertEquals(fs.end(), 0)
        self.assertEquals(fs.padding(), '')
        self.assertEquals(fs.dirname(), '')
        self.assertEquals(fs.basename(), 'file')
        self.assertEquals(fs.extension(), '')
        self.assertEquals(str(fs), "file")

    def testEmptyBasename(self):
        seq = FileSequence("/path/to/1-5#.exr")
        self.assertEquals(seq.basename(), "")
        self.assertEquals(len(seq), 5)
        self.assertEquals(seq.padding(), '#')

        seq = FileSequence("/path/to/1-5%04d.exr")
        self.assertEquals(seq.basename(), "")
        self.assertEquals(len(seq), 5)
        self.assertEquals(seq.padding(), '%04d')

    def test_yield_sequences_in_list(self):
        paths = [
            '/path/to/file.5.png',
            '/path/to/file.1.exr',
            '/path/to/file.2.exr',
            '/path/to/file.3.exr',
            '/path/to/.cruft.file',
            '/path/to/.cruft',
            '/path/to/file2.exr',
            '/path/to/file2.7zip',
            '/path/to/file.2.7zip',
            '/path/to/file.3.7zip',
            '/path/to/file.4.7zip',
            '/path/to/file.4.mp4',
            '', # empty path test
            "mixed_seqs/file5.ext",
            "mixed_seqs/file20.ext",
            "mixed_seqs/file30.ext",
            "mixed_seqs/no_ext",
            "mixed_seqs/no_ext.200",
            "mixed_seqs/no_ext.300",
            "mixed_seqs/no_ext_10",
            "mixed_seqs/not_a_seq.ext",
            "mixed_seqs/seq.0001.ext",
            "mixed_seqs/seq.0002.ext",
            "mixed_seqs/seq.0003.ext",
            "mixed_seqs/seq2a.1.ext",
            "mixed_seqs/seq2a.2.ext",
            "mixed_seqs/seq2a.3.ext",
            "/path/to/file4-4.exr",
            "/path/to/file4-5.exr",
            "/path/to/file--4.exr",
            "path/01.exr",
            "path/02.exr",
            "path/03.exr",
            "path/001.file",
            "path/002.file",
            "path/003.file",
            "path/0001.jpg",
            "path/0002.jpg",
            "path/0003.jpg",
        ]
        actual = set(str(fs) for fs in FileSequence.yield_sequences_in_list(paths))
        expected = set([
            '/path/to/file2@.7zip',
            '/path/to/file.1-3@.exr',
            '/path/to/file.2-4@.7zip',
            '/path/to/file2@.exr',
            '/path/to/file.4@.mp4',
            '/path/to/.cruft.file',
            '/path/to/.cruft',
            '/path/to/file.5@.png',
            "mixed_seqs/file5,20,30@.ext",
            "mixed_seqs/seq2a.1-3@.ext",
            "mixed_seqs/seq.1-3#.ext",
            "mixed_seqs/not_a_seq.ext",
            "mixed_seqs/no_ext",
            "mixed_seqs/no_ext_10@@",
            "mixed_seqs/no_ext.200",
            "mixed_seqs/no_ext.300",
            '/path/to/file4-5,-4@@.exr',
            '/path/to/file--4@@.exr',
            'path/1-3@@.exr',
            'path/1-3@@@.file',
            'path/1-3#.jpg',
        ])
        self.assertEquals(actual, expected)

    def testIgnoreFrameSetStrings(self):
        for char in "xy:,".split():
            fs = FileSequence("/path/to/file{0}1-1x1#.exr".format(char))
            self.assertEquals(fs.basename(), "file{0}".format(char))
            self.assertEquals(fs.start(), 1)
            self.assertEquals(fs.end(), 1)
            self.assertEquals(fs.padding(), '#')
            self.assertEquals(str(fs), "/path/to/file{0}1-1x1#.exr".format(char))

class TestFindSequencesOnDisk(TestBase):

    def testFindSequencesOnDisk(self):
        seqs = findSequencesOnDisk("seq")
        self.assertEquals(6, len(seqs))

        known = set([
            "seq/bar1000-1002,1004-1006#.exr",
            "seq/foo.1-5#.exr",
            "seq/foo.1-5#.jpg",
            "seq/foo.debug.1-5#.exr",
            "seq/foo_1#.exr",
            "seq/1-3#.exr",
        ])
        found = set([str(s) for s in seqs])
        self.assertEqualPaths(found, known)

    def testNegSequencesOnDisk(self):
        seqs = findSequencesOnDisk("seqneg")
        self.assertEquals(1, len(seqs))

    def testFindSequenceOnDiskNegative(self):
        seqs = findSequencesOnDisk("seqneg")
        self.assertEquals("seqneg/bar.-1-1#.exr", str(seqs[0]))
        self.assertEquals("seqneg/bar.-001.exr", seqs[0].frame(-1))
        self.assertEquals("seqneg/bar.-1001.exr", seqs[0].frame(-1001))
        self.assertEquals("seqneg/bar.-10011.exr", seqs[0].frame(-10011))
        self.assertEquals("seqneg/bar.1000.exr", seqs[0].frame(1000))

    def testFindSequencesOnDiskSkipHiddenFiles(self):
        seqs = findSequencesOnDisk("seqhidden")
        self.assertEquals(3, len(seqs))

        known = set(self.toNormpaths([
            "seqhidden/bar1000-1002,1004-1006#.exr",
            "seqhidden/foo.1-5#.exr",
            "seqhidden/foo.1-5#.jpg",
        ]))
        found = set(self.toNormpaths([str(s) for s in seqs]))
        self.assertEqual(known, found)
        self.assertFalse(known.difference(found))

    def testFindSequencesOnDiskIncludeHiddenFiles(self):
        seqs = findSequencesOnDisk("seqhidden", include_hidden=True)
        self.assertEquals(7, len(seqs))

        known = set([
            "seqhidden/bar1000-1002,1004-1006#.exr",
            "seqhidden/.bar1000-1002,1004-1006#.exr",
            "seqhidden/foo.1-5#.exr",
            "seqhidden/.foo.1-5#.exr",
            "seqhidden/foo.1-5#.jpg",
            "seqhidden/.foo.1-5#.jpg",
            "seqhidden/.hidden",
        ])
        found = set([str(s) for s in seqs])
        self.assertEqualPaths(known, found)

    def testCrossPlatformPathSep(self):
        expected = set([
            "seqsubdirs/sub1/1-3#.exr",
            "seqsubdirs/sub1/bar1000-1002,1004-1006#.exr",
            "seqsubdirs/sub1/foo.1-5#.exr",
            "seqsubdirs/sub1/foo.1-5#.jpg",
            "seqsubdirs/sub1/foo.debug.1-5#.exr",
            "seqsubdirs/sub1/foo_1#.exr",
        ])

        import ntpath 
        _join = os.path.join
        os.path.join = ntpath.join

        try:
            self.assertEqual(os.path.join('a', 'b'), 'a\\b')
            seqs = findSequencesOnDisk("seqsubdirs/sub1")

            self.assertEquals(len(expected), len(seqs))

            actual = set(str(s) for s in seqs)
            self.assertEqual(actual, expected)

        finally:
            os.path.join = _join

class TestFindSequenceOnDisk(TestBase):

    def testFindSequenceOnDisk(self):
        tests = [
            ("seq/bar#.exr", "seq/bar1000-1002,1004-1006#.exr"),
            ("seq/foo.#.exr", "seq/foo.1-5#.exr"),
            ("seq/foo.#.jpg", "seq/foo.1-5#.jpg"),
            ("seq/foo.0002.jpg", "seq/foo.1-5#.jpg"),
            ("seq/foo.#.exr", "seq/foo.1-5#.exr"),
            ("seq/foo.debug.#.exr", "seq/foo.debug.1-5#.exr"),
            ("seq/#.exr", "seq/1-3#.exr"),
            ("seq/bar1001.exr", "seq/bar1001.exr"),
            ("seq/foo_0001.exr", "seq/foo_0001.exr"),
        ]

        for pattern, expected in tests:
            seq = findSequenceOnDisk(pattern)
            self.assertTrue(isinstance(seq, FileSequence))
            actual = str(seq)
            self.assertEqual(actual, expected)

    def testCrossPlatformPathSep(self):
        tests = [
            ("seq/bar#.exr", "seq\\bar1000-1002,1004-1006#.exr"),
            ("seq/foo.#.exr", "seq\\foo.1-5#.exr"),
            ("seq/foo.#.jpg", "seq\\foo.1-5#.jpg"),
            ("seq/foo.0002.jpg", "seq\\foo.1-5#.jpg"),
            ("seq/foo.#.exr", "seq\\foo.1-5#.exr"),
            ("seq/foo.debug.#.exr", "seq\\foo.debug.1-5#.exr"),
            ("seq/#.exr", "seq\\1-3#.exr"),
            ("seq/bar1001.exr", "seq/bar1001.exr"),
            ("seq/foo_0001.exr", "seq/foo_0001.exr"),
        ]

        import ntpath 
        _path = os.path
        os.path = ntpath

        try:
            self.assertEqual(os.path.join('a', 'b'), 'a\\b')

            for pattern, expected in tests:
                seq = findSequenceOnDisk(pattern)
                self.assertTrue(isinstance(seq, FileSequence))
                actual = str(seq)
                self.assertEqual(actual, expected) 

        finally:
            os.path = _path

    def testPaddingMatch(self):
        tests = [
            ("mixed/seq.#.ext", "mixed/seq.-1-5#.ext"),
            ("mixed/seq.@@.ext", "mixed/seq.-1-5@@.ext"),
            ("mixed/seq.@@@@@.ext", "mixed/seq.-1-5@@@@@.ext"),
            ("mixed/seq.@.ext", None),
            ("mixed/seq.%04d.ext", "mixed/seq.-1-5#.ext"),
            ("mixed/seq.%02d.ext", "mixed/seq.-1-5@@.ext"),
            ("mixed/seq.%05d.ext", "mixed/seq.-1-5@@@@@.ext"),
            ("mixed/seq.%01d.ext", None),
        ]

        for pattern, expected in tests:
            if expected is None:
                with self.assertRaises(FileSeqException):
                    findSequenceOnDisk(pattern, strictPadding=True)
                continue
            
            seq = findSequenceOnDisk(pattern, strictPadding=True)
            self.assertTrue(isinstance(seq, FileSequence))
            
            actual = str(seq)
            self.assertEqual(actual, expected)


class TestPaddingFunctions(unittest.TestCase):
    """
    Test functions that help deal with padding on file sequences.
    """

    def testgetPaddingChars(self):
        """
        Ensure that we're getting back the proper padding characters.
        :return: None
        """
        self.assertEqual(getPaddingChars(0), '@')
        self.assertEqual(getPaddingChars(1), '@')
        self.assertEqual(getPaddingChars(2), '@@')
        self.assertEqual(getPaddingChars(3), '@@@')
        self.assertEqual(getPaddingChars(4), '#')
        self.assertEqual(getPaddingChars(8), '##')

    def testgetPaddingNum(self):
        """
        Ensure that we're getting back the proper padding number.
        :return: None
        """
        self.assertEqual(getPaddingNum('@'), 1)
        self.assertEqual(getPaddingNum('@@'), 2)
        self.assertEqual(getPaddingNum('@@@'), 3)
        self.assertEqual(getPaddingNum('#'), 4)
        self.assertEqual(getPaddingNum('##'), 8)
        self.assertEqual(getPaddingNum('#@'), 5)
        self.assertEqual(getPaddingNum('##@@'), 10)

        self.assertEqual(getPaddingNum('%01d'), 1)
        self.assertEqual(getPaddingNum('%1d'), 1)
        self.assertEqual(getPaddingNum('%04d'), 4)
        self.assertEqual(getPaddingNum('%10d'), 10)

        allPossibleChars = [s for s in string.printable if s not in PAD_MAP.keys()]
        for char in allPossibleChars:
            self.assertRaises(ValueError, getPaddingNum, char)
            self.assertRaises(ValueError, getPaddingNum, '#{}'.format(char))
            self.assertRaises(ValueError, getPaddingNum, '@{}'.format(char))

        allPossibleChars = [s for s in string.printable if s not in PAD_MAP.keys() and s not in string.digits]
        for char in allPossibleChars:
            self.assertRaises(ValueError, getPaddingNum, '%{}d'.format(char))

    def testPadFrameRange(self):
        self.assertEqual(padFrameRange('1', 6), '000001')
        self.assertEqual(padFrameRange('-1', 6), '-000001')
        self.assertEqual(padFrameRange('1-100', 6), '000001-000100')
        self.assertEqual(padFrameRange('-1-100', 6), '-000001-000100')
        self.assertEqual(padFrameRange('-1--100', 6), '-000001--000100')
        self.assertEqual(padFrameRange('1--100', 6), '000001--000100')
        self.assertEqual(padFrameRange('1-100x2', 6), '000001-000100x2')
        self.assertEqual(padFrameRange('-1-100x2', 6), '-000001-000100x2')
        self.assertEqual(padFrameRange('-1--100x2', 6), '-000001--000100x2')
        self.assertEqual(padFrameRange('1--100x2', 6), '000001--000100x2')
        self.assertEqual(padFrameRange('1--100x2', 5), '00001--00100x2')
        self.assertEqual(padFrameRange('1--100x2', 4), '0001--0100x2')
        self.assertEqual(padFrameRange('1--100x2', 3), '001--100x2')
        self.assertEqual(padFrameRange('1--100x2', 2), '01--100x2')
        self.assertEqual(padFrameRange('1--100x2', 1), '1--100x2')
        self.assertEqual(padFrameRange('1--100x2', 0), '1--100x2')
        self.assertEqual(padFrameRange('1--100x2', -1), '1--100x2')

if __name__ == '__main__':
    unittest.main(verbosity=1)





#! /usr/bin/env python
"""
constants - General constants of use to fileseq operations.
"""

import re

# The max frame count of a FrameSet before a MaxSizeException
# exception is raised
MAX_FRAME_SIZE = 10000000
    
PAD_MAP = {"#": 4, "@": 1}

# Regular expression for matching a file sequence string.
# Example: /film/shot/renders/bilbo_bty.1-100#.exr
# Example: /film/shot/renders/bilbo_bty.1-100@.exr
# Example: /film/shot/renders/bilbo_bty.1-100@@@@#.exr
# Example: /film/shot/renders/bilbo_bty.1-100%04d.exr
SPLIT_PATTERN = r"((?:[-\d][-:,xy\d]*)?)([{0}]+|%\d+d)".format(''.join(PAD_MAP.keys()))
SPLIT_RE = re.compile(SPLIT_PATTERN)

# Regular expression pattern for matching padding against a printf syntax
# padding string E.g. %04d
PRINTF_SYNTAX_PADDING_PATTERN = r"%(\d+)d"
PRINTF_SYNTAX_PADDING_RE = re.compile(PRINTF_SYNTAX_PADDING_PATTERN)

# Regular expression pattern for matching file names on disk.
DISK_PATTERN = r"^((?:.*[/\\])?)(?:(.*?)(-?\d+)?((?:\.[^.]*)?))?$"
DISK_RE = re.compile(DISK_PATTERN)

# Regular expression pattern for matching frame set strings.
# Examples: '1' or '1-100', '1-100x5', '1-100:5', '1-100y5', '1,2', etc.
FRANGE_PATTERN = r"^(-?\d+)(?:-(-?\d+)(?:([:xy]{1})(\d+))?)?$"
FRANGE_RE = re.compile(FRANGE_PATTERN)

# Regular expression for padding a frame range.
PAD_PATTERN = r"(-?)(\d+)(?:(-)(-?)(\d+)(?:([:xy]{1})(\d+))?)?"
PAD_RE = re.compile(PAD_PATTERN)
